class NfDictionaryTryGetResults:
    def __init__(self):
        self.__key_exists = False

        self.__value = None

    def __get_key_exists(self) -> bool:
        key_exists = self.__key_exists

        return key_exists

    def __set_key_exists(
        self,
        key_exists: bool,
    ):
        self.__key_exists = key_exists

    def __get_value(self):
        value = self.__value

        return value

    def __set_value(self, value):
        self.__value = value

    key_exists = property(
        fget=__get_key_exists,
        fset=__set_key_exists,
    )

    value = property(
        fget=__get_value,
        fset=__set_value,
    )



from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_element_types import (
    EaElementTypes,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)
from pandas import DataFrame


def shift_convention_objects_to_classes(
    content_universe: NfEaComUniverses,
    output_universe: NfEaComUniverses,
) -> None:
    __run_input_checks()

    __run_operation(
        content_universe=content_universe,
        output_universe=output_universe,
    )


def __run_operation(
    content_universe: NfEaComUniverses,
    output_universe: NfEaComUniverses,
) -> None:
    log_message(
        message="CONVENTION SHIFT OPERATION: Shift objects to classes - started",
    )

    content_collections_dictionary = (
        content_universe.nf_ea_com_registry.dictionary_of_collections
    )

    for (
        content_collection_type,
        content_collection_table,
    ) in (
        content_collections_dictionary.items()
    ):
        __process_content_collection(
            content_collection_type=content_collection_type,
            content_collection_table=content_collection_table,
            output_universe=output_universe,
        )

    log_message(
        message="CONVENTION SHIFT OPERATION: Shift objects to classes - finished",
    )


def __run_input_checks():
    pass


def __process_content_collection(
    content_collection_type: NfEaComCollectionTypes,
    content_collection_table: DataFrame,
    output_universe: NfEaComUniverses,
) -> None:
    output_collections_dictionary = (
        output_universe.nf_ea_com_registry.dictionary_of_collections
    )

    if (
        content_collection_type
        == NfEaComCollectionTypes.EA_CLASSIFIERS
    ):
        __convert_objects_to_classes(
            output_collections_dictionary=output_collections_dictionary,
            ea_classifiers=content_collection_table,
        )

    else:
        output_collections_dictionary[
            content_collection_type
        ] = content_collection_table


def __convert_objects_to_classes(
    output_collections_dictionary: dict,
    ea_classifiers: DataFrame,
) -> None:
    ea_object_type_column_name = (
        NfEaComColumnTypes.ELEMENTS_EA_OBJECT_TYPE.column_name
    )

    number_of_objects_before_conversion = len(
        ea_classifiers[
            ea_classifiers[
                ea_object_type_column_name
            ]
            == EaElementTypes.OBJECT.type_name
        ].index,
    )
    # TODO: replaced to address warning from pandas : FutureWarning: A value is trying to be set on a copy of a DataFrame or Series through chained assignment using an inplace method.
    #   The behavior will change in pandas 3.0. This inplace method will never work because the intermediate object on which we are setting values always behaves as a copy.
    # ea_classifiers[
    #     ea_object_type_column_name
    # ].replace(
    #     to_replace=EaElementTypes.OBJECT.type_name,
    #     value=EaElementTypes.CLASS.type_name,
    #     inplace=True,
    # )

    ea_classifiers[
        ea_object_type_column_name
    ] = ea_classifiers[
        ea_object_type_column_name
    ].replace(
        to_replace=EaElementTypes.OBJECT.type_name,
        value=EaElementTypes.CLASS.type_name,
    )

    number_of_objects_after_conversion = len(
        ea_classifiers[
            ea_classifiers[
                ea_object_type_column_name
            ]
            == EaElementTypes.OBJECT.type_name
        ].index,
    )

    number_of_objects_converted_to_classes = (
        number_of_objects_before_conversion
        - number_of_objects_after_conversion
    )

    log_message(
        message="CONVENTION SHIFT OPERATION: Shift objects to classes - Number of objects converted: "
        + str(
            number_of_objects_converted_to_classes,
        ),
    )

    output_collections_dictionary[
        NfEaComCollectionTypes.EA_CLASSIFIERS
    ] = ea_classifiers



"""Template for pipeline orchestrator."""

PIPELINE_ORCHESTRATOR_IMPORT_TEMPLATE = """from bclearer_pipelines.{domain_name}.b_source.{pipeline_name}.orchestrators.thin_slices.{thin_slice_name}_orchestrator import (
    orchestrate_{thin_slice_name},
)
"""

PIPELINE_ORCHESTRATOR_FUNCTION_START = """

def orchestrate_{pipeline_name}():
    __run_contained_bie_pipeline_components()


def __run_contained_bie_pipeline_components() -> (
    None
):
"""

ORCHESTRATE_THIN_SLICE_CALL = """    orchestrate_{thin_slice_name}()
"""

EMPTY_FUNCTION_BODY = """    pass
"""



from enum import auto, unique

from bclearer_core.common_knowledge.operation_types import (
    OperationTypes,
)


@unique
class ConventionShiftOperationTypes(
    OperationTypes,
):
    OBJECTS_TO_CLASSES = auto()
    UML_NAME_TO_NAMED_OBJECT = auto()
    ATTRIBUTE_NAME_TO_NAMED_OBJECT = (
        auto()
    )
    GENERALISE_NAMES = auto()
    SEPARATE_STANDARD_NAMES_AND_INSTANCES = (
        auto()
    )
    SEPARATE_BESPOKE_NAMES_AND_INSTANCES = (
        auto()
    )
    SEPARATE_STANDARD_INSTANCES_AND_EXEMPLARS = (
        auto()
    )
    SEPARATE_BESPOKE_INSTANCES_AND_EXEMPLARS = (
        auto()
    )
    UNIVERSE_TO_SEMANTICALLY_GROUNDED_DIGITALISATION_LEVEL = (
        auto()
    )

    def __operation_name(self) -> str:
        operation_name = (
            operation_name_mapping[self]
        )

        return operation_name

    operation_name = property(
        fget=__operation_name,
    )


operation_name_mapping = {
    ConventionShiftOperationTypes.OBJECTS_TO_CLASSES: "shift_from_objects_to_classes",
    ConventionShiftOperationTypes.UML_NAME_TO_NAMED_OBJECT: "shift_from_uml_name_to_named_object",
    ConventionShiftOperationTypes.ATTRIBUTE_NAME_TO_NAMED_OBJECT: "shift_from_attribute_name_to_name_object",
    ConventionShiftOperationTypes.GENERALISE_NAMES: "generalise_names",
    ConventionShiftOperationTypes.SEPARATE_STANDARD_NAMES_AND_INSTANCES: "separate_standard_names_and_instances",
    ConventionShiftOperationTypes.SEPARATE_BESPOKE_NAMES_AND_INSTANCES: "separate_bespoke_names_and_instances",
    ConventionShiftOperationTypes.SEPARATE_STANDARD_INSTANCES_AND_EXEMPLARS: "separate_standard_instances_and_exemplars",
    ConventionShiftOperationTypes.SEPARATE_BESPOKE_INSTANCES_AND_EXEMPLARS: "separate_bespoke_instances_and_exemplars",
    ConventionShiftOperationTypes.UNIVERSE_TO_SEMANTICALLY_GROUNDED_DIGITALISATION_LEVEL: "shift_universe_to_semantically_grounded",
}



from enum import Enum, auto


class LoggingInspectionLevelBEnums(
    Enum
):
    NOT_SET = auto()

    VERBOSE = auto()

    INFO = auto()

    WARNING = auto()

    ERROR = auto()

    CRITICAL = auto()

    def __enum_level_weight(
        self,
    ) -> str:
        enum_level_weight = (
            enum_level_weight_mapping[
                self
            ]
        )

        return enum_level_weight

    enum_level_weight = property(
        fget=__enum_level_weight
    )

    def __enum_color(
        self,
    ) -> str:
        enum_color = (
            enum_color_mapping[
                self
            ]
        )

        return enum_color

    enum_color = property(
        fget=__enum_color
    )


enum_level_weight_mapping = {
    LoggingInspectionLevelBEnums.NOT_SET: int(),
    LoggingInspectionLevelBEnums.VERBOSE: 1,
    LoggingInspectionLevelBEnums.INFO: 2,
    LoggingInspectionLevelBEnums.WARNING: 3,
    LoggingInspectionLevelBEnums.ERROR: 4,
    LoggingInspectionLevelBEnums.CRITICAL: 5,
}

# ANSI color codes for console output
enum_color_mapping = {
    LoggingInspectionLevelBEnums.NOT_SET: "",
    LoggingInspectionLevelBEnums.VERBOSE: "\033[90m",  # Gray
    LoggingInspectionLevelBEnums.INFO: "\033[36m",     # Cyan
    LoggingInspectionLevelBEnums.WARNING: "\033[93m",  # Amber (Bright Yellow)
    LoggingInspectionLevelBEnums.ERROR: "\033[91m\033[1m",    # Bold Red
    LoggingInspectionLevelBEnums.CRITICAL: "\033[95m\033[1m",  # Bold Magenta
}

# ANSI reset code
COLOR_RESET = "\033[0m"



"""Template for application runner."""

APPLICATION_RUNNER_TEMPLATE = """from bclearer_orchestration_services.b_app_runner_service.b_application_runner import (
    run_b_application,
)
from bclearer_pipelines.{domain_name}.b_source.app_runners.runners.{domain_name}_b_clearer_pipelines_runner import (
    run_{domain_name}_b_clearer_pipelines,
)


def run_{domain_name}_b_clearer_pipeline_b_application() -> (
    None
):
    run_b_application(
        app_startup_method=run_{domain_name}_b_clearer_pipelines
    )
"""



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.b_evolve.common.universes_merge_registers import (
    UniversesMergeRegisters,
)
from bclearer_core.substages.operations.b_evolve.content_operations.merge_universes.nf_ea_com_collection_processes.collection_type_compliation_constants import (
    LIST_OF_COLLECTION_TYPES_OF_OBJECTS_WITHOUT_EA_GUIDS,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)
from pandas import concat


def concat_universe_collections(
    collection_type: NfEaComCollectionTypes,
    universe_merge_register: UniversesMergeRegisters,
    output_universe: NfEaComUniverses,
):
    if not collection_type:
        log_message(
            message="CONTENT OPERATION: Concat universes - concatenating collections - No Collection type - Skipped",
        )
        return

    log_message(
        message="CONTENT OPERATION: Concat universes - concatenating collections - "
        + collection_type.collection_name,
    )

    primary_collection = universe_merge_register.primary_universe.nf_ea_com_registry.dictionary_of_collections.get(
        collection_type
    )
    aligned_collection = universe_merge_register.aligned_universe.nf_ea_com_registry.dictionary_of_collections.get(
        collection_type
    )

    if (
        collection_type
        in LIST_OF_COLLECTION_TYPES_OF_OBJECTS_WITHOUT_EA_GUIDS
    ):
        concatenated_collection = (
            primary_collection.copy()
        )
    else:
        # Filter out empty DataFrames before concatenation
        collections_to_concat = [
            df
            for df in [
                primary_collection,
                aligned_collection,
            ]
            if not df.empty
        ]

        if collections_to_concat:
            concatenated_collection = concat(
                collections_to_concat
            )
        else:
            concatenated_collection = (
                primary_collection
            )

    concatenated_collection.reset_index(
        drop=True,
        inplace=True,
    )

    if (
        NfColumnTypes.NF_UUIDS.column_name
        in concatenated_collection.columns
    ):
        concatenated_collection.drop_duplicates(
            subset=NfColumnTypes.NF_UUIDS.column_name,
            inplace=True,
        )

    output_universe.nf_ea_com_registry.dictionary_of_collections[
        collection_type
    ] = concatenated_collection


# TODO: patched, needs to be replaced
#
# def concat_universe_collections(
#     collection_type: NfEaComCollectionTypes,
#     universe_merge_register: UniversesMergeRegisters,
#     output_universe: NfEaComUniverses,
# ):
#     if not collection_type:
#         log_message(
#             message="CONTENT OPERATION: Concat universes - concatenating collections - No Collection type - Skipped",
#         )
#
#         return
#
#     log_message(
#         message="CONTENT OPERATION: Concat universes - concatenating collections - "
#         + collection_type.collection_name,
#     )
#
#     if (
#         collection_type
#         in LIST_OF_COLLECTION_TYPES_OF_OBJECTS_WITHOUT_EA_GUIDS
#     ):
#         concatenated_collection = universe_merge_register.primary_universe.nf_ea_com_registry.dictionary_of_collections[
#             collection_type
#         ].copy()
#
#     else:
#         concatenated_collection = concat(
#             [
#                 universe_merge_register.primary_universe.nf_ea_com_registry.dictionary_of_collections[
#                     collection_type
#                 ],
#                 universe_merge_register.aligned_universe.nf_ea_com_registry.dictionary_of_collections[
#                     collection_type
#                 ],
#             ],
#         )
#
#     concatenated_collection.reset_index(
#         drop=True,
#         inplace=True,
#     )
#
#     if (
#         NfColumnTypes.NF_UUIDS.column_name
#         in concatenated_collection.columns
#     ):
#         concatenated_collection.drop_duplicates(
#             subset=NfColumnTypes.NF_UUIDS.column_name,
#             inplace=True,
#         )
#
#     output_universe.nf_ea_com_registry.dictionary_of_collections[
#         collection_type
#     ] = concatenated_collection



from bclearer_core.objects.b_units.b_units import BUnits
from bclearer_core.objects.b_units.b_units_with_object_passing_and_returning import (
    BUnitsWithObjectPassingAndReturning,
)

__all__ = [
    "BUnits",
    "BUnitsWithObjectPassingAndReturning",
]



from bclearer_core.infrastructure.nf.objects.registries.nf_verse_registries import (
    NfVerseRegistries,
)


class NfUniverseRegistries(
    NfVerseRegistries
):
    """Universe-level registries for verses."""



"""Helpers for determining the current operating system as an enum."""

import platform

from bclearer_core.infrastructure.session.common.enums.operating_system_kind import (
    OperatingSystemKind,
)


# TODO: Harmonise file and public method name (and internal variable names)
def get_operating_system_kind() -> (
    OperatingSystemKind
):
    system_name = platform.system()

    if system_name == "Windows":
        return (
            OperatingSystemKind.WINDOWS
        )

    if system_name == "Linux":
        return OperatingSystemKind.LINUX

    if system_name == "Darwin":
        return OperatingSystemKind.MACOS

    return OperatingSystemKind.UNKNOWN



from bclearer_core.infrastructure.session.common.enums.operating_system_kind import (
    OperatingSystemKind,
)
from bclearer_core.infrastructure.session.common.platform_type_getter import (
    get_operating_system_kind,
)
from bclearer_core.infrastructure.session.operating_system.bie_operating_system_types import (
    BieOperatingSystemTypes,
)


def get_bie_operating_system_type() -> (
    BieOperatingSystemTypes
):
    operating_system_kind = (
        get_operating_system_kind()
    )

    if (
        operating_system_kind
        is OperatingSystemKind.WINDOWS
    ):
        return (
            BieOperatingSystemTypes.BIE_WINDOWS
        )

    if (
        operating_system_kind
        is OperatingSystemKind.LINUX
    ):
        return (
            BieOperatingSystemTypes.BIE_LINUX
        )

    if (
        operating_system_kind
        is OperatingSystemKind.MACOS
    ):
        return (
            BieOperatingSystemTypes.BIE_MAC_OS
        )

    return (
        BieOperatingSystemTypes.NOT_SET
    )



"""Template for a simple startup script to run e2e tests."""

STARTUP_SCRIPT_TEMPLATE = """#!/usr/bin/env bash
set -euo pipefail

SCRIPT_DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"
cd "${SCRIPT_DIR}"

if ! command -v uv >/dev/null 2>&1; then
    echo "uv is not installed. Install uv to run this script."
    exit 1
fi

uv venv "${SCRIPT_DIR}/.venv"
source "${SCRIPT_DIR}/.venv/bin/activate"
uv sync --active --no-install-project

if [ -z "${BCLEARER_REPO_ROOT:-}" ]; then
    echo "BCLEARER_REPO_ROOT is not set. Export it to the repo root containing libraries/."
    exit 1
fi

LEGACY_PIPELINES_PATH="${SCRIPT_DIR}/../bclearer_pipelines"

export PYTHONPATH="${LEGACY_PIPELINES_PATH}:${BCLEARER_REPO_ROOT}/libraries/core:${BCLEARER_REPO_ROOT}/libraries/orchestration_services:${BCLEARER_REPO_ROOT}/libraries/interop_services${PYTHONPATH:+:${PYTHONPATH}}"

python -m pytest "${SCRIPT_DIR}/tests/universal/e2e"
"""



# TODO Circular reference to be fixed - super class to be added
# TODO: fix setup for configurations with single called imports - cannot change after setup
class BieIdConfigurations:
    EXPORT_INT_VALUE_COLUMNS = False



from bclearer_core.bie.top.bie_objects import (
    BieObjects,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.common_knowledge.bie_infrastructure_types import (
    BieInfrastructureTypes,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.objects.bie_ids import (
    BieIds,
)


class BieInfrastructureObjects(
    BieObjects
):
    def __init__(
        self,
        bie_id: BieIds,
        base_hr_name: str,
        bie_infrastructure_type: BieInfrastructureTypes,
    ):
        super().__init__(
            bie_id,
            base_hr_name=base_hr_name,
            bie_type=bie_infrastructure_type,
        )



from bclearer_core.infrastructure.session.common.enums.operating_system_kind import (
    OperatingSystemKind,
)
from bclearer_core.infrastructure.session.common.platform_type_getter import (
    get_operating_system_kind,
)
from bclearer_core.infrastructure.session.user.bie_os_user_profile_types import (
    BieOsUserProfileTypes,
)


def get_bie_os_user_profile_type() -> (
    BieOsUserProfileTypes
):
    operating_system_kind = (
        get_operating_system_kind()
    )

    if (
        operating_system_kind
        is OperatingSystemKind.WINDOWS
    ):
        return (
            BieOsUserProfileTypes.BIE_WINDOWS_USER_PROFILE
        )

    if (
        operating_system_kind
        is OperatingSystemKind.LINUX
    ):
        return (
            BieOsUserProfileTypes.BIE_LINUX_USER_PROFILE
        )

    if (
        operating_system_kind
        is OperatingSystemKind.MACOS
    ):
        return (
            BieOsUserProfileTypes.BIE_MAC_OS_USER_PROFILE
        )

    return BieOsUserProfileTypes.NOT_SET



from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.column_types.ea_t.ea_t_xref_column_types import (
    EaTXrefColumnTypes,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)
from pandas import DataFrame


def remove_attributes(
    content_universe: NfEaComUniverses,
    adjustment_universe: NfEaComUniverses,
    output_universe: NfEaComUniverses,
) -> None:
    __run_input_checks()

    __run_operation(
        content_universe=content_universe,
        adjustment_universe=adjustment_universe,
        output_universe=output_universe,
    )


def __run_operation(
    content_universe: NfEaComUniverses,
    adjustment_universe: NfEaComUniverses,
    output_universe: NfEaComUniverses,
) -> None:
    log_message(
        message="ADJUSTMENT OPERATION: Remove attributes - started",
    )

    content_collections_dictionary = (
        content_universe.nf_ea_com_registry.dictionary_of_collections
    )

    for (
        content_collection_type,
        content_collection_table,
    ) in (
        content_collections_dictionary.items()
    ):
        __process_content_collection(
            content_collection_type=content_collection_type,
            content_collection_table=content_collection_table,
            adjustment_universe=adjustment_universe,
            output_universe=output_universe,
        )

    # TODO: Report errors

    log_message(
        message="ADJUSTMENT OPERATION: Remove attributes - finished",
    )


def __run_input_checks():
    pass


def __process_content_collection(
    content_collection_type: NfEaComCollectionTypes,
    content_collection_table: DataFrame,
    adjustment_universe: NfEaComUniverses,
    output_universe: NfEaComUniverses,
) -> None:
    output_collections_dictionary = (
        output_universe.nf_ea_com_registry.dictionary_of_collections
    )

    list_of_attribute_ea_guids_to_be_removed = __get_list_of_attribute_ea_guids_to_be_removed(
        adjustment_universe=adjustment_universe,
    )

    if (
        content_collection_type
        == NfEaComCollectionTypes.EA_ATTRIBUTES
    ):
        output_collections_dictionary[
            content_collection_type
        ] = __get_filtered_collection(
            list_of_attribute_ea_guids_to_be_removed=list_of_attribute_ea_guids_to_be_removed,
            content_collection_table=content_collection_table,
            ea_guid_column_name=NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name,
        )

    elif (
        content_collection_type
        == NfEaComCollectionTypes.STEREOTYPE_USAGE
    ):
        output_collections_dictionary[
            content_collection_type
        ] = __get_filtered_collection(
            list_of_attribute_ea_guids_to_be_removed=list_of_attribute_ea_guids_to_be_removed,
            content_collection_table=content_collection_table,
            ea_guid_column_name=EaTXrefColumnTypes.T_XREF_CLIENT_EA_GUIDS.nf_column_name,
        )

    else:
        output_collections_dictionary[
            content_collection_type
        ] = content_collection_table


def __get_list_of_attribute_ea_guids_to_be_removed(
    adjustment_universe: NfEaComUniverses,
) -> list:
    adjustment_collections_dictionary = (
        adjustment_universe.nf_ea_com_registry.dictionary_of_collections
    )

    table_of_attributes_to_be_removed = adjustment_collections_dictionary[
        NfEaComCollectionTypes.EA_ATTRIBUTES
    ]

    list_of_attribute_ea_guids_to_be_removed = table_of_attributes_to_be_removed[
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
    ].to_list()

    return list_of_attribute_ea_guids_to_be_removed


def __get_filtered_collection(
    list_of_attribute_ea_guids_to_be_removed: list,
    content_collection_table: DataFrame,
    ea_guid_column_name: str,
) -> DataFrame:
    filtered_collection = content_collection_table[
        ~content_collection_table[
            ea_guid_column_name
        ].isin(
            list_of_attribute_ea_guids_to_be_removed,
        )
    ]

    return filtered_collection



from bclearer_core.configurations.bie_configurations.bie_configurations import (
    BieConfigurations,
)


class BieCytoscapeInspectionConfigurations(
    BieConfigurations
):
    def __init__(self):
        super().__init__()

    ENABLE_CYTOSCAPE_GRAPH_INSPECTION = (
        True
    )

    CYTOSCAPE_IS_RUNNING = False

    CYTOSCAPE_MAXIMUM_ALLOCATED_RAM_MEMORY = (
        "4"
    )

    CYTOSCAPE_WAIT_TIME = 30

    CYTOSCAPE_NUMBER_OF_WAIT_INTERVALS = (
        5
    )



from enum import auto

from bclearer_orchestration_services.identification_services.b_identity_ecosystem.common_knowledge.bie_enums import (
    BieEnums,
)


# TODO: Rename to a better name. Maybe SessionObjectTableNames??
class SessionObjectColumnNames(
    BieEnums
):
    NOT_SET = auto()

    APP_RUN_SESSIONS = auto()

    OPERATING_SYSTEMS = auto()

    OS_USER_PROFILES = auto()

    OS_USER_PROFILE_SESSIONS = auto()



from bclearer_core.bie.infrastructure.types.bie_common_column_names import (
    BieCommonColumnNames,
)
from bclearer_core.infrastructure.session.os_user_session.enums.os_user_profile_object_column_names import (
    OsUserProfileObjectColumnNames,
)
from bclearer_core.infrastructure.session.user.os_user_profile_objects import (
    OsUserProfileObjects,
)
from bclearer_interop_services.b_dictionary_service.table_as_dictionary_service.table_as_dictionary_to_dataframe_converter import (
    convert_table_as_dictionary_to_dataframe,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.common_knowledge.bie_column_names import (
    BieColumnNames,
)
from pandas import DataFrame


def create_os_user_profile_objects_table() -> (
    DataFrame
):
    table_as_dictionary = (
        __get_os_user_profile_object_table_as_dictionary()
    )

    table_as_dataframe = convert_table_as_dictionary_to_dataframe(
        table_as_dictionary=table_as_dictionary
    )

    return table_as_dataframe.astype(
        str
    )


def __get_os_user_profile_object_table_as_dictionary() -> (
    dict
):
    os_user_profile_object = (
        OsUserProfileObjects()
    )

    os_user_profile_row = {
        BieColumnNames.BIE_IDS.b_enum_item_name: os_user_profile_object.bie_id,
        BieCommonColumnNames.BASE_HR_NAMES.b_enum_item_name: os_user_profile_object.base_hr_name,
        BieCommonColumnNames.BIE_INFRASTRUCTURE_TYPE_IDS.b_enum_item_name: os_user_profile_object.bie_type.item_bie_identity,
        OsUserProfileObjectColumnNames.USER_NAMES.b_enum_item_name: os_user_profile_object.user_name,
        OsUserProfileObjectColumnNames.USER_SIDS.b_enum_item_name: str(
            os_user_profile_object.user_sid
        ),
        OsUserProfileObjectColumnNames.DOMAIN_NAMES.b_enum_item_name: os_user_profile_object.domain_name,
        OsUserProfileObjectColumnNames.ACCOUNT_TYPES.b_enum_item_name: os_user_profile_object.account_type,
    }

    table_as_dictionary = {
        0: os_user_profile_row
    }

    return table_as_dictionary



from bclearer_core.configurations.workspaces.factories.workspace_configuration_factory import (
    create_workspace_configuration,
)
from bclearer_interop_services.file_system_service.objects.folders import (
    Folders,
)
#TODO: needs to be created/fixed
from bclearer_core.configurations.b_app_configurations import (
    BAppConfigurations,
)


def create_and_set_this_b_app_run_workspace_configuration(
    this_b_app_run_nf_workspace: Folders,
    this_b_app_run_workspace_type_prefix: (
        str | None
    ) = None,
    this_b_app_run_workspace_type_suffix: (
        str | None
    ) = None,
) -> None:
    this_b_app_run_workspace_configuration = create_workspace_configuration(
        parent_workspace=this_b_app_run_nf_workspace,
        workspace_type_prefix=this_b_app_run_workspace_type_prefix,
        workspace_type_suffix=this_b_app_run_workspace_type_suffix,
    )

    BAppConfigurations.THIS_B_APP_RUN_WORKSPACE_CONFIGURATION = this_b_app_run_workspace_configuration



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from bclearer_core.configurations.datastructure.logging_inspection_level_b_enums import (
    LoggingInspectionLevelBEnums,
)
from bclearer_core.objects.b_units.b_units import BUnits
from bclearer_core.objects.universes.universes import Universes
from bclearer_orchestration_services.reporting_service.reporters.inspection_message_logger import (
    log_inspection_message,
)

TUniverse = TypeVar("TUniverse", bound=Universes)


class BUnitsWithObjectPassingAndReturning(
    BUnits,
    Generic[TUniverse],
    ABC,
):
    """
    Generic B-Unit that accepts and operates on a Universe object.

    This pattern enables pipeline steps to:
    - Receive a Universe containing pipeline state and data
    - Access registries and configuration through the Universe
    - Mutate the Universe by adding/modifying data in registers
    - Pass the modified Universe to the next B-Unit
    """

    def __init__(
        self,
        input_object: TUniverse,
    ) -> None:
        super().__init__()

        self.input_object: TUniverse = input_object

    def run(self) -> None:
        log_inspection_message(
            message=f"Running B-Unit: {self.__class__.__name__}",
            logging_inspection_level_b_enum=LoggingInspectionLevelBEnums.INFO,
        )

        self.b_unit_process_function()

    def b_unit_process_function(
        self,
    ) -> ():
        self._b_unit_process_function()

    @abstractmethod
    def _b_unit_process_function(
        self,
    ) -> None:
        pass



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from bclearer_core.configurations.datastructure.logging_inspection_level_b_enums import LoggingInspectionLevelBEnums
from bclearer_orchestration_services.reporting_service.reporters.inspection_message_logger import log_inspection_message

T = \
    TypeVar('T')

class Registers(ABC, Generic[T]):
    """
    Generic base class for register objects that store typed content.
    
    A Register is the lowest level in the hierarchy, storing a single piece
    of content (e.g., a DataFrame, a graph database, etc.).
    
    Type Parameters:
        T: The type of content this register will store.
        
    Attributes:
        owning_registry: Back-reference to the parent Registry object.
        name: The name/identifier for this register.
        _content: The stored content of type T.
    """
    
    def __init__(
            self,
            owning_registry,
            name: str,
            content: T | None = None):
        super().__init__()

        self.owning_registry = \
            owning_registry
        
        self.name = \
            name
        
        self._content: T | None = \
            content
        
        self._content_type: type = \
            T

    def has_content(
            self)\
            -> bool:
        """
        Check if this register contains content.
        
        :returns: True if content is set, False otherwise.
        """
        return \
            self._content is not None

    @abstractmethod
    def get_model(self):
        """
        Get metadata/schema information about the content.
        
        This should return a dictionary or object describing the structure
        of the content without returning the content itself.
        
        :returns: Metadata about the content.
        :raises NotImplementedError: Must be implemented by subclasses.
        """
        raise \
            NotImplementedError()

    def get_content(
            self)\
            -> T:
        """
        Retrieve the content from this register.
        
        :returns: The stored content of type T.
        :raises ValueError: If content is not set.
        """
        if self.has_content() is False:
            raise \
                ValueError(
                    f"Content for register '{self.name}' is not set.")
        
        # COMMENT: This is here so there is no typing error. get_content does not understand
        # that has_content ensures _content is not None.
        assert \
            self._content is not None

        return \
            self._content
    
    def get_content_type(
            self) \
            -> type:
        """
        Get the type of content this register stores.
        :returns: The type of the content.
        """
        return \
            self._content_type

    def set_content(
        self, 
        content: T):
        """
        Store content in this register.
        
        If content already exists, it will be overwritten with a warning.
        
        :param content: The content to store.
        """
        if self.has_content():
            log_inspection_message(
                message=f"Warning: Overwriting content for register '{self.name}'.",
                logging_inspection_level_b_enum=LoggingInspectionLevelBEnums.WARNING)
            
        self._content = \
            content

    @abstractmethod
    def export_to_disk(
            self,
            base_folder_path: str):
        """
        Export the register's content to disk.
        
        :param base_folder_path: The base folder path for export.
        :raises NotImplementedError: Must be implemented by subclasses.
        """
        raise \
            NotImplementedError()
    
    def log_register_state(
            self) \
            -> None:
        """
        Log the state of this register.
        
        Default implementation logs basic information about the register.
        Override this method for custom logging behavior specific to content type.
        """
        has_content = \
            self.has_content()
        
        status = \
            "has content" if has_content else "is empty"
        
        log_inspection_message(
            logging_inspection_level_b_enum=LoggingInspectionLevelBEnums.INFO,
            message=f"Register '{self.name}' {status}.")


from enum import auto

from bclearer_orchestration_services.identification_services.b_identity_ecosystem.common_knowledge.bie_enums import (
    BieEnums,
)


class AppRunSessionObjectColumnNames(
    BieEnums
):
    NOT_SET = auto()

    START_DATE_TIMES = auto()

    OWNING_OS_USER_PROFILE_SESSION_BIE_IDS = (
        auto()
    )



from bclearer_core.constants.standard_constants import (
    DEFAULT_NULL_VALUE,
)
from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.common.ea_guid_from_nf_uuid_creator import (
    create_ea_guid_from_nf_uuid,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_element_types import (
    EaElementTypes,
)
from bclearer_orchestration_services.identification_services.uuid_service.uuid_helpers.uuid_factory import (
    create_new_uuid,
)


def add_new_classifier_to_dictionary(
    new_classifier_dictionary: dict,
    package_nf_uuid: str,
    ea_element_type: EaElementTypes,
    class_name: str = "",
    elements_classifier: str = DEFAULT_NULL_VALUE,
) -> str:
    class_nf_uuid = create_new_uuid()

    class_ea_guid = (
        create_ea_guid_from_nf_uuid(
            nf_uuid=class_nf_uuid,
        )
    )

    class_row_dictionary = {
        NfColumnTypes.NF_UUIDS.column_name: class_nf_uuid,
        NfEaComColumnTypes.ELEMENTS_CLASSIFIER.column_name: elements_classifier,
        NfEaComColumnTypes.CLASSIFIERS_ALL_COMPONENT_EA_ATTRIBUTES.column_name: [],
        NfEaComColumnTypes.ELEMENTS_EA_OBJECT_TYPE.column_name: ea_element_type.type_name,
        NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name: [],
        NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS.column_name: [],
        NfEaComColumnTypes.ELEMENTS_CONTAINED_EA_DIAGRAMS.column_name: [],
        NfEaComColumnTypes.ELEMENTS_CONTAINED_EA_CLASSIFIERS.column_name: [],
        NfEaComColumnTypes.PACKAGEABLE_OBJECTS_PARENT_EA_ELEMENT.column_name: package_nf_uuid,
        NfEaComColumnTypes.STEREOTYPEABLE_OBJECTS_EA_OBJECT_STEREOTYPES.column_name: [],
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name: class_name,
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name: class_ea_guid,
    }

    new_classifier_dictionary[
        class_nf_uuid
    ] = class_row_dictionary

    return class_nf_uuid



from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)


def get_intersection_of_dependency_and_association_linked(
    nf_ea_com_universe: NfEaComUniverses,
    linked_by_dependency_nf_uuid: str,
    linked_by_association_nf_uuid: str,
) -> str:
    connected_by_dependency_set = get_connected_classifiers(
        nf_ea_com_universe=nf_ea_com_universe,
        classifier_nf_uuid=linked_by_dependency_nf_uuid,
        connection_type=EaConnectorTypes.DEPENDENCY,
        connected_end_type=NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS,
    )

    connected_by_association_set = get_connected_classifiers(
        nf_ea_com_universe=nf_ea_com_universe,
        classifier_nf_uuid=linked_by_association_nf_uuid,
        connection_type=EaConnectorTypes.ASSOCIATION,
        connected_end_type=NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS,
    )

    intersection = list(
        connected_by_dependency_set
        & connected_by_association_set,
    )

    if len(intersection) == 1:
        return intersection[0]

    raise ValueError(
        "Too many objects in intersection",
    )


def get_connected_classifiers(
    nf_ea_com_universe: NfEaComUniverses,
    classifier_nf_uuid: str,
    connection_type: EaConnectorTypes,
    connected_end_type: NfEaComColumnTypes,
) -> set:
    ea_connectors = nf_ea_com_universe.nf_ea_com_registry.dictionary_of_collections[
        NfEaComCollectionTypes.EA_CONNECTORS
    ]

    connections = ea_connectors[
        ea_connectors[
            connected_end_type.column_name
        ]
        == classifier_nf_uuid
    ]

    connections_by_type = connections[
        connections[
            NfEaComColumnTypes.CONNECTORS_ELEMENT_TYPE_NAME.column_name
        ]
        == connection_type.type_name
    ]

    if (
        connected_end_type
        == NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS
    ):
        connected_classifiers = set(
            connections_by_type[
                NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name
            ],
        )

    elif (
        connected_end_type
        == NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS
    ):
        connected_classifiers = set(
            connections_by_type[
                NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS.column_name
            ],
        )

    else:
        raise TypeError(
            "Invalid connector end column: "
            + connected_end_type.column_name,
        )

    return connected_classifiers



from bclearer_core.configurations.bie_configurations.bie_configurations import (
    BieConfigurations,
)


class BieGraphMlInspectionConfigurations(
    BieConfigurations
):
    def __init__(self):
        super().__init__()

    PERSIST_GRAPH_ML_FOR_INSPECTION = (
        True
    )



from bclearer_core.common_knowledge.digitialisation_level_stereotype_matched_ea_objects import (
    DigitalisationLevelStereotypeMatchedEaObjects,
)
from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.b_evolve.common.digitalisation_level_stereotypes_usages_getter import (
    get_digitalisation_level_stereotypes_usages,
)
from bclearer_core.substages.operations.common.new_ea_objects_dictionary_creator import (
    create_new_ea_objects_dictionary,
)
from bclearer_core.substages.operations.common.nf_ea_com_universe_updater import (
    update_nf_ea_com_universe_with_dictionary,
)
from bclearer_core.substages.operations.common.nf_uuid_from_ea_guid_from_collection_getter import (
    get_nf_uuid_from_ea_guid_from_collection,
)
from bclearer_core.substages.operations.common.stereotype_adder import (
    add_new_stereotype_usage_to_dictionary,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from pandas import DataFrame


def replace_digitalisation_level_stereotype(
    nf_ea_com_universe: NfEaComUniverses,
    digitalisation_level_stereotype: DigitalisationLevelStereotypeMatchedEaObjects,
) -> None:
    current_digitalisation_level_stereotype_usages = get_digitalisation_level_stereotypes_usages(
        nf_ea_com_universe=nf_ea_com_universe,
    )

    __delete_current_digitalisation_level_stereotype_usages(
        nf_ea_com_universe=nf_ea_com_universe,
    )

    __add_replacement_digitalisation_level_stereotype_usages(
        nf_ea_com_universe=nf_ea_com_universe,
        digitalisation_level_stereotype=digitalisation_level_stereotype,
        digitalisation_level_stereotype_usages_to_replace=current_digitalisation_level_stereotype_usages,
    )


def __delete_current_digitalisation_level_stereotype_usages(
    nf_ea_com_universe: NfEaComUniverses,
):
    digitalisation_level_stereotype_ea_guids = (
        DigitalisationLevelStereotypeMatchedEaObjects.get_ea_guids()
    )

    ea_stereotypes = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_stereotypes()
    )

    digitalisation_level_stereotypes = ea_stereotypes[
        ea_stereotypes[
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
        ].isin(
            digitalisation_level_stereotype_ea_guids,
        )
    ]

    digitalisation_level_stereotypes_nf_uuids = set(
        digitalisation_level_stereotypes[
            NfColumnTypes.NF_UUIDS.column_name
        ],
    )

    ea_stereotype_usages = nf_ea_com_universe.nf_ea_com_registry.dictionary_of_collections[
        NfEaComCollectionTypes.STEREOTYPE_USAGE
    ]

    stereotype_usages_to_keep = ea_stereotype_usages[
        ~ea_stereotype_usages[
            "stereotype_nf_uuids"
        ].isin(
            digitalisation_level_stereotypes_nf_uuids,
        )
    ]

    nf_ea_com_universe.nf_ea_com_registry.dictionary_of_collections[
        NfEaComCollectionTypes.STEREOTYPE_USAGE
    ] = stereotype_usages_to_keep


def __add_replacement_digitalisation_level_stereotype_usages(
    nf_ea_com_universe: NfEaComUniverses,
    digitalisation_level_stereotype: DigitalisationLevelStereotypeMatchedEaObjects,
    digitalisation_level_stereotype_usages_to_replace: DataFrame,
):
    new_ea_objects_dictionary = (
        create_new_ea_objects_dictionary()
    )

    update_nf_ea_com_universe_with_dictionary(
        nf_ea_com_universe=nf_ea_com_universe,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
    )

    digitalisation_level_stereotype_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_STEREOTYPES,
        ea_guid=digitalisation_level_stereotype.ea_guid,
    )

    classifier_digitalisation_level_stereotyped_nf_uuids = set(
        digitalisation_level_stereotype_usages_to_replace[
            NfEaComColumnTypes.STEREOTYPE_CLIENT_NF_UUIDS.column_name
        ],
    )

    for (
        ea_classifier_nf_uuid
    ) in classifier_digitalisation_level_stereotyped_nf_uuids:
        add_new_stereotype_usage_to_dictionary(
            new_stereotype_usage_dictionary=new_ea_objects_dictionary[
                NfEaComCollectionTypes.STEREOTYPE_USAGE
            ],
            client_nf_uuid=ea_classifier_nf_uuid,
            client_collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
            stereotype_nf_uuid=digitalisation_level_stereotype_nf_uuid,
        )

    update_nf_ea_com_universe_with_dictionary(
        nf_ea_com_universe=nf_ea_com_universe,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
    )



from bclearer_orchestration_services.identification_services.b_identity_ecosystem.common_knowledge.bie_infrastructure_types import (
    BieInfrastructureTypes,
)


class BieAppRunSessionTypes(
    BieInfrastructureTypes
):
    pass



"""Windows-specific helpers for machine identification."""

import ctypes

from bclearer_core.infrastructure.session.common.session_config import (
    WINDOWS_BUILTIN_ADMINISTRATORS_SID,
    WINDOWS_SECURITY_MAX_SID_SIZE,
)


# TODO: Harmonise file and public method name (and internal variable names)
def get_windows_machine_sid() -> str:
    sid = ctypes.create_string_buffer(
        WINDOWS_SECURITY_MAX_SID_SIZE
    )
    sid_size = ctypes.c_ulong(
        WINDOWS_SECURITY_MAX_SID_SIZE
    )

    if not ctypes.windll.advapi32.CreateWellKnownSid(
        WINDOWS_BUILTIN_ADMINISTRATORS_SID,
        None,
        sid,
        ctypes.byref(sid_size),
    ):
        raise ctypes.WinError()

    sid_string = ctypes.c_wchar_p()

    if not ctypes.windll.advapi32.ConvertSidToStringSidW(
        sid, ctypes.byref(sid_string)
    ):
        raise ctypes.WinError()

    machine_sid = (
        sid_string.value.rsplit("-", 1)[
            0
        ]
    )
    ctypes.windll.kernel32.LocalFree(
        sid_string
    )

    return machine_sid



from bclearer_core.configurations.universe_modification_operation_configurations import (
    UniverseModificationOperationConfigurations,
)
from bclearer_core.substages.operations.b_evolve.universe_modification_operations.universe_modification_operation_substages import (
    UniverseModificationOperationsSubstages,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.session.orchestrators.ea_tools_session_managers import (
    EaToolsSessionManagers,
)


def run_universe_modification_operations_substage(
    ea_tools_session_manager: EaToolsSessionManagers,
    content_universe: NfEaComUniverses,
    universe_modification_operation_configuration: UniverseModificationOperationConfigurations,
) -> NfEaComUniverses:
    with UniverseModificationOperationsSubstages(
        ea_tools_session_manager=ea_tools_session_manager,
        universe_modification_operation_configuration=universe_modification_operation_configuration,
        content_universe=content_universe,
    ) as universe_modification_operations_substage:
        universe_modification_operations_substage_output_universe = (
            universe_modification_operations_substage.run()
        )

        return universe_modification_operations_substage_output_universe



"""Template for thin slice orchestrator."""

THIN_SLICE_ORCHESTRATOR_IMPORT_TEMPLATE = """from bclearer_pipelines.{domain_name}.b_source.{pipeline_name}.orchestrators.stages.{pipeline_name}_{stage_name}_orchestrator import (
    orchestrate_{pipeline_name}_{stage_name},
)
"""

THIN_SLICE_ORCHESTRATOR_FUNCTION_START = """

def orchestrate_{thin_slice_name}():
    __run_contained_bie_pipeline_components()


def __run_contained_bie_pipeline_components() -> (
    None
):
"""

ORCHESTRATE_STAGE_CALL = """    orchestrate_{pipeline_name}_{stage_name}()
"""

EMPTY_FUNCTION_BODY = """    pass
"""



from bclearer_core.configurations.b_clearer_configurations.b_clearer_configurations import (
    BClearerConfigurations,
)


class BClearerDomainConfigurations(
    BClearerConfigurations
):
    pass



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from pandas import DataFrame


def has_classifier_ea_guid_attribute_type_name_and_value(
    content_universe: NfEaComUniverses,
    classifier_ea_guid: str,
    attribute_type_name: str,
    attribute_value: str,
) -> bool:
    content_universe_ea_classifiers = (
        content_universe.nf_ea_com_registry.get_ea_classifiers()
    )

    classifier_list_of_attribute_nf_uuids = __get_classifier_list_of_attribute_nf_uuids(
        content_universe_ea_classifiers=content_universe_ea_classifiers,
        classifier_ea_guid=classifier_ea_guid,
    )

    classifier_ea_guid_has_attribute_type_name_and_value = (
        False
    )

    content_universe_ea_attributes = (
        content_universe.nf_ea_com_registry.get_ea_attributes()
    )

    for (
        attribute_nf_uuid
    ) in classifier_list_of_attribute_nf_uuids:
        classifier_ea_guid_has_attribute_type_name_and_value = __check_attribute_and_type_is_a_match(
            content_universe_ea_attributes=content_universe_ea_attributes,
            content_universe_ea_classifiers=content_universe_ea_classifiers,
            attribute_value=attribute_value,
            attribute_type_name=attribute_type_name,
            attribute_nf_uuid=attribute_nf_uuid,
        )

        if classifier_ea_guid_has_attribute_type_name_and_value:
            return True

    return classifier_ea_guid_has_attribute_type_name_and_value


def __get_classifier_list_of_attribute_nf_uuids(
    content_universe_ea_classifiers: DataFrame,
    classifier_ea_guid: str,
) -> list:
    ea_guid_column_name = (
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
    )

    all_component_ea_attributes_column_name = (
        NfEaComColumnTypes.CLASSIFIERS_ALL_COMPONENT_EA_ATTRIBUTES.column_name
    )

    classifier_list_of_attribute_nf_uuids = content_universe_ea_classifiers[
        content_universe_ea_classifiers[
            ea_guid_column_name
        ]
        == classifier_ea_guid
    ][
        all_component_ea_attributes_column_name
    ].tolist()[
        0
    ]

    return classifier_list_of_attribute_nf_uuids


def __check_attribute_and_type_is_a_match(
    content_universe_ea_attributes: DataFrame,
    attribute_nf_uuid: str,
    content_universe_ea_classifiers: DataFrame,
    attribute_value: str,
    attribute_type_name: str,
) -> bool:
    nf_uuids_column_name = (
        NfColumnTypes.NF_UUIDS.column_name
    )

    ea_object_name_column_name = (
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name
    )

    classifying_ea_classifier_column_name = (
        NfEaComColumnTypes.ELEMENT_COMPONENTS_CLASSIFYING_EA_CLASSIFIER.column_name
    )

    current_attribute_dataframe_slice = content_universe_ea_attributes[
        content_universe_ea_attributes[
            nf_uuids_column_name
        ]
        == attribute_nf_uuid
    ]

    if (
        current_attribute_dataframe_slice.empty
    ):
        return False

    current_attribute_value = current_attribute_dataframe_slice[
        ea_object_name_column_name
    ].values[
        0
    ]

    current_attribute_type_nf_uuid_string = current_attribute_dataframe_slice[
        classifying_ea_classifier_column_name
    ].values[
        0
    ]

    current_attribute_type_name_result = content_universe_ea_classifiers[
        content_universe_ea_classifiers[
            nf_uuids_column_name
        ]
        == current_attribute_type_nf_uuid_string
    ][
        ea_object_name_column_name
    ]

    if (
        not current_attribute_type_name_result.empty
    ):
        current_attribute_type_name = current_attribute_type_name_result.values[
            0
        ]

        if (
            attribute_value
            == current_attribute_value
            and attribute_type_name
            == current_attribute_type_name
        ):
            return True

    return False



from bclearer_core.configurations.adjustment_operations_substage_configurations import (
    AdjustmentOperationsSubstageConfigurations,
)
from bclearer_core.substages.operations.b_evolve.adjustment_operations.adjustment_operations_substages import (
    AdjustmentOperationsSubstages,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.session.orchestrators.ea_tools_session_managers import (
    EaToolsSessionManagers,
)


def run_adjustment_operations_substage(
    ea_tools_session_manager: EaToolsSessionManagers,
    content_universe: NfEaComUniverses,
    adjustment_operations_substage_configuration: AdjustmentOperationsSubstageConfigurations,
) -> NfEaComUniverses:
    with AdjustmentOperationsSubstages(
        ea_tools_session_manager=ea_tools_session_manager,
        content_universe=content_universe,
        adjustment_operations_substage_configuration=adjustment_operations_substage_configuration,
    ) as adjustment_operations_substage:
        adjustment_operations_substage_output_universe = (
            adjustment_operations_substage.run()
        )

        return adjustment_operations_substage_output_universe



from bclearer_core.common_knowledge.matched_objects import (
    MatchedEaObjects,
)


class UmlNameToNamedObjectConfigurationObjects:
    def __init__(
        self,
        matched_package: MatchedEaObjects,
        matched_naming_space: MatchedEaObjects,
    ):
        self.matched_package = (
            matched_package
        )

        self.matched_naming_space = (
            matched_naming_space
        )

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



from bclearer_core.configurations.load_ea_model_configurations import (
    LoadEaModelConfigurations,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.session.orchestrators.ea_tools_session_managers import (
    EaToolsSessionManagers,
)
from bclearer_interop_services.ea_interop_service.session.processes.creators.new_nf_ea_com_universe_using_file_creator import (
    create_new_nf_ea_com_universe_using_file,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)
from bclearer_orchestration_services.reporting_service.wrappers.run_and_log_function_wrapper import (
    run_and_log_function,
)
from bclearer_core.services.resources_service.processes.resource_file_getter import (
    get_resource_file,
)


@run_and_log_function
def load_ea_model_to_content_universe(
    ea_tools_session_manager: EaToolsSessionManagers,
    load_ea_model_configuration: LoadEaModelConfigurations,
) -> NfEaComUniverses:
    log_message(
        message="CONTENT OPERATION: Load universe - "
        + load_ea_model_configuration.short_name
        + " - started",
    )

    ea_repository_file = get_resource_file(
        resource_namespace=load_ea_model_configuration.resource_namespace,
        resource_name=load_ea_model_configuration.resource_name,
    )

    output_universe = create_new_nf_ea_com_universe_using_file(
        ea_tools_session_manager=ea_tools_session_manager,
        ea_repository_file=ea_repository_file,
        short_name=load_ea_model_configuration.short_name,
    )

    log_message(
        message="CONTENT OPERATION: Load universe - "
        + load_ea_model_configuration.short_name
        + " - finished",
    )

    return output_universe



"""Configuration constants for infrastructure session helpers."""

from pathlib import Path

WINDOWS_SECURITY_MAX_SID_SIZE = 68
WINDOWS_BUILTIN_ADMINISTRATORS_SID = 26

LINUX_MACHINE_ID_PATHS = (
    Path("/etc/machine-id"),
    Path("/var/lib/dbus/machine-id"),
)
LINUX_MACHINE_ID_FALLBACK_PREFIX = (
    "linux-mac-"
)

MACOS_IOREG_COMMAND = [
    "ioreg",
    "-rd1",
    "-c",
    "IOPlatformExpertDevice",
]
MACOS_IOREG_UUID_PATTERN = (
    r'"IOPlatformUUID" = "([^"]+)"'
)
MACOS_MACHINE_FALLBACK_PREFIX = (
    "macos-mac-"
)

UNKNOWN_MACHINE_FALLBACK_PREFIX = (
    "unknown-machine-"
)

UNIX_DEFAULT_DOMAIN_NAME = "LOCAL"
UNIX_ADMIN_ACCOUNT_TYPE = (
    "Administrator"
)
UNIX_STANDARD_ACCOUNT_TYPE = "User"



from abc import ABC
from enum import Enum
from typing import Any, Dict, Generic, TypeVar
import os

from bclearer_interop_services.file_system_service.objects.folders import \
    Folders
from bclearer_core.objects.registries.registries import Registries

TRegistry = \
    TypeVar('TRegistry', bound=Registries)


class Universes(ABC, Generic[TRegistry]):
    """
    Generic base class for Universe objects that manage collections of registries.
    
    A Universe is the top-level container in the hierarchy, managing multiple
    Registries (which in turn manage Registers). It also maintains a configuration
    registry for pipeline-level settings.
    
    Type Parameters:
        TRegistry: The type of registries this universe will contain.
        
    Attributes:
        _registries: Dictionary mapping enum keys to registry instances.
        _configuration_registry: Dictionary mapping enum keys to configuration values.
    """
    
    def __init__(
            self):
        super().__init__()

        self._registries: Dict[Enum, TRegistry] = \
            {}
        
        self._configuration_registry: Dict[Enum, Any] = \
            {}
        
    @property
    def registries(
            self) \
            -> Dict[Enum, TRegistry]:
        """
        Get all registries in this universe.
        
        :returns: Dictionary mapping enum keys to registry instances.
        """
        return \
            self._registries
    
    @property
    def configuration_registry(
            self) \
            -> Dict[Enum, Any]:
        """
        Get the configuration registry.
        
        :returns: Dictionary mapping enum keys to configuration values.
        """
        return \
            self._configuration_registry
    
    def get_registry(
            self,
            name: Enum) \
            -> TRegistry:
        """
        Retrieve a registry by its enum name.
        
        :param name: The enum key for the registry.
        :returns: The registry instance.
        :raises ValueError: If registry not found.
        """
        if name not in self._registries:
            raise \
                ValueError(
                    f"Registry '{name}' not found in {self.__class__.__name__}")
        
        return \
            self._registries[name]

    def set_registry(
            self,
            name: Enum,
            registry: TRegistry) \
            -> None:
        """
        Store a registry by its enum name.
        
        Also sets the registry as an attribute for direct access.
        
        :param name: The enum key for the registry.
        :param registry: The registry instance to store.
        """
        self._registries[name] = \
            registry
        
        # Also set as attribute for direct access (e.g., universe.sap_registry)
        setattr(
            self, 
            name.value.lower(), 
            registry)
    
    def has_registry(
            self,
            name: Enum) \
            -> bool:
        """
        Check if a registry exists by name.
        
        :param name: The enum key to check.
        :returns: True if registry exists, False otherwise.
        """
        return \
            name in self._registries
    
    def get_all_registry_names(
            self) \
            -> list[Enum]:
        """
        Get a list of all registry names in this universe.
        
        :returns: List of enum keys for all registries.
        """
        return \
            list(self._registries.keys())
    
    def get_configuration(
            self,
            name: Enum,
            default: Any = None) \
            -> Any:
        """
        Get a configuration value by name.
        
        :param name: The enum key for the configuration.
        :param default: Default value if configuration not found.
        :returns: The configuration value or default.
        """
        return \
            self._configuration_registry.get(
                name,
                default)
    
    def set_configuration(
            self,
            name: Enum,
            value: Any) \
            -> None:
        """
        Set a configuration value by name.
        
        :param name: The enum key for the configuration.
        :param value: The configuration value to store.
        """
        self._configuration_registry[name] = \
            value
    
    def has_configuration(
            self,
            name: Enum) \
            -> bool:
        """
        Check if a configuration exists by name.
        
        :param name: The enum key to check.
        :returns: True if configuration exists, False otherwise.
        """
        return \
            name in self._configuration_registry
    
    def export_to_disk(
            self,
            base_folder_path: str) \
            -> None:
        """
        Export all registries to disk.
        
        Default implementation creates a subfolder structure and delegates
        to each registry's export_to_disk method. Override this method if
        custom export logic is needed.
        
        :param base_folder_path: The base folder path for exports.
        """
        universe_name = \
            self.__class__.__name__.lower()
        
        base_folder = \
            Folders(
                absolute_path_string=base_folder_path)
        
        export_folder = \
            base_folder.absolute_path \
            / universe_name \
            / "registries"
        
        if not export_folder.exists():
            os.makedirs(
                export_folder,
                exist_ok=True)
        
        # Delegate to all registries
        for registry in self._registries.values():
            registry.export_to_disk(
                folder_path=str(export_folder))
    
    def log_universe_state(
            self) \
            -> None:
        """
        Log the state of all registries in this universe.
        
        Default implementation delegates to each registry's log_registry_state method.
        Override this method if custom logging behavior is needed.
        """
        from bclearer_core.configurations.datastructure.logging_inspection_level_b_enums import \
            LoggingInspectionLevelBEnums
        from bclearer_orchestration_services.reporting_service.reporters.inspection_message_logger import \
            log_inspection_message
        
        universe_name = \
            self.__class__.__name__
        
        log_inspection_message(
            logging_inspection_level_b_enum=LoggingInspectionLevelBEnums.INFO,
            message=f"Universe '{universe_name}' contains {len(self._registries)} registr(y/ies).")
        
        # Delegate to all registries
        for registry in self._registries.values():
            registry.log_registry_state()



from bclearer_core.infrastructure.nf.objects.registers.nf_universe_registers import (
    NfUniverseRegisters,
)


class NfDomainUniverseRegisters(
    NfUniverseRegisters
):
    def __init__(self) -> None:
        super().__init__()



from bclearer_core.infrastructure.nf.objects.registries.nf_universe_registries import (
    NfUniverseRegistries,
)


class NfDomainUniverseRegistries(
    NfUniverseRegistries
):
    def __init__(self) -> None:
        super().__init__()



from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from pandas import DataFrame


def update_nf_ea_com_universe_with_dictionary(
    nf_ea_com_universe: NfEaComUniverses,
    new_ea_objects_dictionary: dict,
) -> None:
    for (
        collection_type,
        new_collection_as_dictionary,
    ) in (
        new_ea_objects_dictionary.items()
    ):
        __update_collection_if_required(
            nf_ea_com_universe=nf_ea_com_universe,
            collection_type=collection_type,
            new_collection_as_dictionary=new_collection_as_dictionary,
        )


def __update_collection_if_required(
    nf_ea_com_universe: NfEaComUniverses,
    collection_type: NfEaComCollectionTypes,
    new_collection_as_dictionary: dict,
):
    if (
        len(
            new_collection_as_dictionary,
        )
        == 0
    ):
        return

    new_collection = DataFrame.from_dict(
        data=new_collection_as_dictionary,
        orient="Index",
    )

    nf_ea_com_universe.nf_ea_com_registry.update(
        collection_type=collection_type,
        new_collection=new_collection,
    )



from bclearer_core.common_knowledge.operation_types import (
    OperationTypes,
)


class OperationConfigurations:
    def __init__(
        self,
        operation_type: OperationTypes,
    ):
        self.operation_type = (
            operation_type
        )

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



from enum import auto, unique

from bclearer_core.common_knowledge.matched_objects import (
    MatchedEaObjects,
)


@unique
class BclearerMatchedEaObjects(
    MatchedEaObjects,
):
    BCLEARER_FOUNDATION_COMMON_RESERVED_NAMES = (
        auto()
    )

    NAMES = auto()

    NAME_INSTANCES = auto()

    NAME_EXEMPLARS = auto()

    CHARACTER_STRINGS = auto()

    MODEL_PACKAGE = auto()

    NAMED_BY_STEREOTYPE = auto()

    NAME_TYPES_INSTANCES_STEREOTYPE = (
        auto()
    )

    EXEMPLIFIED_BY_STEREOTYPE = auto()

    NAME_EXEMPLAR_STEREOTYPE = auto()

    def __object_name(self) -> str:
        object_name = (
            object_name_mapping[self]
        )

        return object_name

    def __ea_guid(self) -> str:
        ea_guid = ea_guid_mapping[self]

        return ea_guid

    object_name = property(
        fget=__object_name,
    )

    ea_guid = property(fget=__ea_guid)


object_name_mapping = {
    BclearerMatchedEaObjects.BCLEARER_FOUNDATION_COMMON_RESERVED_NAMES: "bCLEARer Foundation Common Reserved Names",
    BclearerMatchedEaObjects.NAMES: "Names",
    BclearerMatchedEaObjects.NAME_INSTANCES: "Name Instances",
    BclearerMatchedEaObjects.NAME_EXEMPLARS: "Name Exemplars",
    BclearerMatchedEaObjects.CHARACTER_STRINGS: "Character Strings",
    BclearerMatchedEaObjects.MODEL_PACKAGE: "bCLEARer Foundation Package",
    BclearerMatchedEaObjects.NAMED_BY_STEREOTYPE: "named by",
    BclearerMatchedEaObjects.NAME_TYPES_INSTANCES_STEREOTYPE: "name types-instances",
    BclearerMatchedEaObjects.EXEMPLIFIED_BY_STEREOTYPE: "exemplified by",
    BclearerMatchedEaObjects.NAME_EXEMPLAR_STEREOTYPE: "name exemplar",
}


ea_guid_mapping = {
    BclearerMatchedEaObjects.BCLEARER_FOUNDATION_COMMON_RESERVED_NAMES: "{7fd8c00c-2b46-11eb-86df-983b8f873eed}",
    BclearerMatchedEaObjects.NAMES: "{7fd9834a-2b46-11eb-84e7-983b8f873eed}",
    BclearerMatchedEaObjects.NAME_INSTANCES: "{25E1D82C-87F0-4585-9401-B839DF7F17AB}",
    BclearerMatchedEaObjects.NAME_EXEMPLARS: "{07846B84-EF0C-4596-A54D-6E389DDB247D}",
    BclearerMatchedEaObjects.CHARACTER_STRINGS: "{6054422D-3626-4351-97E3-11EB0852BD2D}",
    BclearerMatchedEaObjects.MODEL_PACKAGE: "{72bb7541-2b46-11eb-96c9-983b8f873eed}",
    BclearerMatchedEaObjects.NAMED_BY_STEREOTYPE: "{7CBE78A4-6CDA-49c4-82CC-AB89A6AF99D9}",
    BclearerMatchedEaObjects.NAME_TYPES_INSTANCES_STEREOTYPE: "{44D7CCD7-3943-4b8f-92A6-67D0BB0BD12F}",
    BclearerMatchedEaObjects.EXEMPLIFIED_BY_STEREOTYPE: "{A8B38219-0B26-46f9-82EC-E7556CCE2120}",
    BclearerMatchedEaObjects.NAME_EXEMPLAR_STEREOTYPE: "{1C49F9D8-5D83-4dcc-B49A-665E759A6EB1}",
}



from enum import auto

from bclearer_orchestration_services.identification_services.b_identity_ecosystem.common_knowledge.bie_enums import (
    BieEnums,
)


# TODO: needs to move to a general area
class BieCommonColumnNames(BieEnums):
    NOT_SET = auto()

    BASE_HR_NAMES = auto()

    BIE_DOMAIN_TYPE_NAMES = auto()

    BIE_SNAPSHOT_TYPE_NAMES = auto()

    BIE_INFRASTRUCTURE_TYPE_IDS = auto()

    BIE_DOMAIN_TYPE_IDS = auto()

    BIE_WHOLE_IDS = auto()

    BIE_PART_IDS = auto()

    BIE_SIGNATURE_SUM_IDS = auto()

    BIE_SNAPSHOT_TYPE_IDS = auto()



from bclearer_core.infrastructure.nf.objects.registers.nf_verse_registers import (
    NfVerseRegisters,
)


class NfUniverseRegisters(
    NfVerseRegisters
):
    """Universe-level verse registers."""



"""Windows-specific helpers for OS user session identifiers."""

import ctypes


def get_windows_user_session_id() -> (
    int
):
    wts_get_active_console_session_id = (
        ctypes.windll.kernel32.WTSGetActiveConsoleSessionId
    )

    return (
        wts_get_active_console_session_id()
    )



import csv
import importlib
import os
import shutil
from os.path import isfile, join
from pathlib import Path

import pyodbc as odbc_library
from bclearer_core.nf.types.collection_types import (
    CollectionTypes,
)
from bclearer_core.nf.types.common_collection_types import (
    CommonCollectionTypes,
)
from bclearer_interop_services.excel_services.excel_facades import ExcelFacades
from bclearer_interop_services.file_system_service.objects.folders import (
    Folders,
)
from bclearer_interop_services.relational_database_services.access_service.access.csv_folder_to_database_loader import (
    load_database_with_table,
)
from pandas import DataFrame, concat

# TODO: make this database agnostic, use a generic database wrapper class


class NfRegistries:
    def __init__(self):
        self.dictionary_of_collections = (
            {}
        )

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass

    def export_dataframes_to_new_database(
        self,
        short_name: str,
        output_folder_name: str,
        database_basename: str,
        add_xlsx_export=True,
    ):
        output_folder = Path(
            output_folder_name,
        )

        output_database_folder = (
            output_folder.joinpath(
                short_name,
                short_name
                + "_"
                + database_basename,
            )
        )

        output_csv_folder = output_database_folder.joinpath(
            "csvs",
        )

        if add_xlsx_export:
            self.__export_dataframes_as_csv_and_xlsx(
                output_csv_folder=output_csv_folder,
                output_database_folder=output_database_folder,
                short_name=short_name,
            )

        else:
            self.export_dataframes(
                short_name=short_name,
                output_folder_name=str(
                    output_csv_folder,
                )
                + os.sep,
            )

        module = importlib.import_module(
            name="bclearer_orchestration_services.resources.templates",
        )

        module_path_string = (
            module.__path__[0]
        )

        resource_full_file_name = (
            os.path.join(
                module_path_string,
                "empty.accdb",
            )
        )

        target_full_file_name = (
            os.path.join(
                output_database_folder,
                short_name
                + "_"
                + database_basename
                + ".accdb",
            )
        )

        shutil.copy(
            src=resource_full_file_name,
            dst=target_full_file_name,
        )

        db_connection_string = (
            "Driver={Microsoft Access Driver (*.mdb, *.accdb)};"
            + "Dbq="
            + target_full_file_name
            + ";"
        )

        db_connection = (
            odbc_library.connect(
                db_connection_string,
                autocommit=True,
            )
        )

        filenames = [
            f
            for f in os.listdir(
                output_csv_folder,
            )
            if isfile(
                join(
                    output_csv_folder,
                    f,
                ),
            )
        ]

        for filename in filenames:
            load_database_with_table(
                db_connection=db_connection,
                table_name=filename,
                csv_folder=Folders(
                    output_csv_folder,
                ),
            )

        db_connection.close()

    def __export_dataframes_as_csv_and_xlsx(
        self,
        output_csv_folder: Path,
        output_database_folder: Path,
        short_name: str,
    ):
        output_csv_folder.mkdir(
            parents=True,
            exist_ok=True,
        )

        output_xlsx_folder = output_database_folder.joinpath(
            "xlsxs",
        )

        output_xlsx_folder.mkdir(
            parents=True,
            exist_ok=True,
        )

        for (
            collection_type,
            dataframe,
        ) in (
            self.dictionary_of_collections.items()
        ):
            NfRegistries.__export_dataframe_as_csv_and_xlsx(
                collection_type=collection_type,
                short_name=short_name,
                output_csv_folder=output_csv_folder,
                dataframe=dataframe,
                output_xlsx_folder=output_xlsx_folder,
            )

    @staticmethod
    def __export_dataframe_as_csv_and_xlsx(
        collection_type: CollectionTypes,
        short_name: str,
        output_csv_folder: Path,
        dataframe: DataFrame,
        output_xlsx_folder: Path,
    ):
        if not isinstance(
            collection_type,
            CollectionTypes,
        ):
            raise TypeError

        collection_name = (
            collection_type.collection_name
        )

        if (
            len(collection_name)
            + len(short_name)
            > 59
        ):
            collection_name = (
                collection_name[
                    0 : 59
                    - len(short_name)
                ]
            )

        csv_filepath = (
            output_csv_folder.joinpath(
                short_name
                + "_"
                + str(collection_name)
                + ".csv",
            )
        )

        dataframe.to_csv(
            str(csv_filepath),
            sep=",",
            quotechar='"',
            index=False,
            quoting=csv.QUOTE_ALL,
            escapechar="\\",
        )

        xlsx_filepath = (
            output_xlsx_folder.joinpath(
                short_name
                + "_"
                + str(collection_name)
                + ".xlsx",
            )
        )

        sheet_name = str(collection_type.collection_name)

        excel_facade = ExcelFacades(str(xlsx_filepath))
        try:
            excel_sheet = excel_facade.workbook.sheet(sheet_name)
        except ValueError:
            excel_sheet = excel_facade.workbook.create_sheet(sheet_name)

        excel_sheet.save_dataframe(
            table=dataframe,
            full_filename=str(xlsx_filepath),
            sheet_name=sheet_name,
        )

    def export_dataframes(
        self,
        short_name: str,
        output_folder_name: str,
    ):
        output_folder = Path(
            output_folder_name,
        )

        output_folder.mkdir(
            parents=True,
            exist_ok=True,
        )

        for (
            collection_type,
            dataframe,
        ) in (
            self.dictionary_of_collections.items()
        ):
            if not isinstance(
                collection_type,
                CollectionTypes,
            ):
                raise TypeError

            dataframe.to_csv(
                output_folder_name
                + short_name
                + "_"
                + str(
                    collection_type.collection_name,
                )
                + ".csv",
                sep=",",
                quotechar='"',
                index=False,
                quoting=csv.QUOTE_ALL,
            )

    def add_table(
        self,
        table: DataFrame,
        collection_type: CollectionTypes,
    ):
        self.dictionary_of_collections[
            collection_type
        ] = table

    def get_collection(
        self,
        collection_type: CollectionTypes,
        collection_factory,
    ):
        if (
            collection_type
            in self.dictionary_of_collections
        ):
            return self.dictionary_of_collections[
                collection_type
            ]

        collection = (
            collection_factory.create()
        )

        self.add_table(
            table=collection,
            collection_type=collection_type,
        )

        return collection

    def replace_collection(
        self,
        collection_type: CollectionTypes,
        collection: DataFrame,
    ):
        self.add_table(
            table=collection,
            collection_type=collection_type,
        )

    def create_or_update_summary_table(
        self,
    ):
        summary_dictionary = dict()

        for (
            collection_type,
            table,
        ) in (
            self.dictionary_of_collections.items()
        ):
            if not isinstance(
                collection_type,
                CollectionTypes,
            ):
                raise TypeError

            summary_dictionary = self.__add_counts_to_dictionary(
                summary_dictionary=summary_dictionary,
                collection_type=collection_type,
                table=table,
            )

        summary_table_columns = [
            "table_names",
            "column_count",
            "row_count",
        ]

        summary_table = DataFrame.from_dict(
            summary_dictionary,
            orient="index",
            columns=summary_table_columns,
        )

        self.dictionary_of_collections[
            CommonCollectionTypes.SUMMARY_TABLE
        ] = summary_table

    def update(
        self,
        collection_type: CollectionTypes,
        new_collection: DataFrame,
    ):
        if (
            collection_type
            not in self.dictionary_of_collections.keys()
        ):
            self.dictionary_of_collections[
                collection_type
            ] = new_collection

            return

        current_collection = self.dictionary_of_collections[
            collection_type
        ]

        merged_collections = concat(
            objs=[
                current_collection,
                new_collection,
            ],
            ignore_index=True,
            verify_integrity=True,
        )

        self.dictionary_of_collections[
            collection_type
        ] = merged_collections

    @staticmethod
    def __add_counts_to_dictionary(
        summary_dictionary: dict,
        collection_type: CollectionTypes,
        table: DataFrame,
    ) -> dict:
        summary_dictionary[
            collection_type
        ] = [
            collection_type.collection_name,
            table.shape[1],
            table.shape[0],
        ]

        return summary_dictionary



"""Linux specific machine identifier helpers."""

import uuid
from pathlib import Path
from typing import Optional

from bclearer_core.infrastructure.session.common.session_config import (
    LINUX_MACHINE_ID_FALLBACK_PREFIX,
    LINUX_MACHINE_ID_PATHS,
)


def _read_machine_id_file(
    path: str | Path,
) -> Optional[str]:
    try:
        value = (
            Path(path)
            .read_text(encoding="utf-8")
            .strip()
        )
    except FileNotFoundError:
        return None

    return value or None


def get_linux_machine_identifier() -> (
    str
):
    for (
        candidate
    ) in LINUX_MACHINE_ID_PATHS:
        machine_id = (
            _read_machine_id_file(
                candidate
            )
        )
        if machine_id:
            return machine_id

    return f"{LINUX_MACHINE_ID_FALLBACK_PREFIX}{uuid.getnode():012x}"



"""Pipeline generator module for creating pipeline structure from configuration."""

import os
import shutil
from pathlib import Path
from typing import (
    Dict,
    List,
    Set,
)

from bclearer_core.pipeline_builder.schema import (
    B_UNIT_TYPE_BARE,
    B_UNIT_TYPE_WITH_OBJECT_PASSING,
    DomainConfig,
    PipelineConfig,
    StageConfig,
    SubStageConfig,
    ThinSliceConfig,
)
from bclearer_core.pipeline_builder.templates import (
    APPLICATION_RUNNER_TEMPLATE,
    B_UNIT_CREATOR_AND_RUNNER_TEMPLATE,
    B_UNIT_IMPORT_TEMPLATE,
    B_UNIT_TEMPLATE,
    B_UNIT_WITH_OBJECT_PASSING_TEMPLATE,
    CONFTEST_TEMPLATE,
    E2E_TEST_TEMPLATE,
    ORCHESTRATE_STAGE_CALL,
    ORCHESTRATE_SUB_STAGE_CALL,
    ORCHESTRATE_THIN_SLICE_CALL,
    PIPELINE_EMPTY_FUNCTION_BODY,
    PIPELINE_IMPORT_TEMPLATE,
    PIPELINE_ORCHESTRATOR_FUNCTION_START,
    PIPELINE_ORCHESTRATOR_IMPORT_TEMPLATE,
    PIPELINE_RUNNER_CALL,
    PIPELINE_RUNNER_TEMPLATE,
    PIPELINES_EMPTY_FUNCTION_BODY,
    PIPELINES_RUNNER_FUNCTION_START,
    PIPELINES_RUNNER_TEMPLATE_HEADER,
    PYPROJECT_TEMPLATE,
    STARTUP_SCRIPT_TEMPLATE,
    STAGE_B_UNIT_CREATOR_IMPORT,
    STAGE_CREATE_AND_RUN_B_UNIT_CALL,
    STAGE_EMPTY_FUNCTION_BODY,
    STAGE_ORCHESTRATOR_FUNCTION_START,
    STAGE_ORCHESTRATOR_HEADER,
    SUB_STAGE_B_UNIT_CREATOR_IMPORT,
    SUB_STAGE_B_UNIT_IMPORT_TEMPLATE,
    SUB_STAGE_CREATE_AND_RUN_B_UNIT_CALL,
    SUB_STAGE_EMPTY_FUNCTION_BODY,
    SUB_STAGE_IMPORT_TEMPLATE,
    SUB_STAGE_ORCHESTRATOR_FUNCTION_START,
    SUB_STAGE_ORCHESTRATOR_HEADER,
    THIN_SLICE_EMPTY_FUNCTION_BODY,
    THIN_SLICE_ORCHESTRATOR_FUNCTION_START,
    THIN_SLICE_ORCHESTRATOR_IMPORT_TEMPLATE,
)


LEGACY_ROOT_DIRNAME = "bclearer_pipelines"
PIPELINE_FOLDER_SUFFIX = "_pipelines"


def _canonical_domain_path(
    output_base_path: Path,
    domain_name: str,
) -> Path:
    """Return the canonical path for a domain pipeline."""
    return output_base_path / f"{domain_name}{PIPELINE_FOLDER_SUFFIX}"


def _legacy_domain_path(
    output_base_path: Path,
    domain_name: str,
) -> Path:
    """Return the legacy compatibility path for a domain pipeline."""
    return output_base_path / LEGACY_ROOT_DIRNAME / domain_name


def _ensure_legacy_view(
    output_base_path: Path,
    domain_name: str,
    canonical_path: Path,
) -> None:
    """
    Ensure the legacy bclearer_pipelines view exists, pointing at the canonical location.
    """
    legacy_root = (
        output_base_path / LEGACY_ROOT_DIRNAME
    )
    legacy_root.mkdir(
        parents=True,
        exist_ok=True,
    )

    legacy_path = (
        legacy_root / domain_name
    )

    if legacy_path.exists():
        try:
            if os.path.samefile(
                legacy_path,
                canonical_path,
            ):
                return
        except FileNotFoundError:
            # Path may have been removed while resolving
            pass

        raise FileExistsError(
            f"Domain path already exists: {legacy_path}"
        )

    try:
        legacy_path.symlink_to(
            canonical_path,
            target_is_directory=True,
        )
    except OSError:
        shutil.copytree(
            canonical_path,
            legacy_path,
        )


def _sync_legacy_view(
    output_base_path: Path,
    domain_name: str,
    canonical_path: Path,
) -> None:
    """
    Refresh the legacy view so it mirrors the canonical pipeline structure.
    """
    legacy_path = _legacy_domain_path(
        output_base_path,
        domain_name,
    )

    if not legacy_path.exists() and not legacy_path.is_symlink():
        _ensure_legacy_view(
            output_base_path,
            domain_name,
            canonical_path,
        )
        return

    try:
        if os.path.samefile(
            legacy_path,
            canonical_path,
        ):
            return
    except FileNotFoundError:
        # Path vanished during check; recreate below
        pass

    if legacy_path.is_symlink():
        legacy_path.unlink()
        try:
            legacy_path.symlink_to(
                canonical_path,
                target_is_directory=True,
            )
            return
        except OSError:
            # Fall through to copytree if symlink creation is not permitted
            pass

    if legacy_path.is_dir():
        shutil.rmtree(legacy_path)
    else:
        legacy_path.unlink()

    shutil.copytree(
        canonical_path,
        legacy_path,
    )


class PipelineUpdater:
    """Updater class for updating existing pipeline structure with new configuration."""

    def __init__(
        self, pipeline_path: str
    ):
        """
        Initialize the PipelineUpdater.

        Args:
            pipeline_path: Path to the existing pipeline directory
        """
        requested_path = Path(
            pipeline_path
        )

        if (
            not requested_path.exists()
        ):
            raise FileNotFoundError(
                f"Pipeline path not found: {pipeline_path}"
            )

        if (
            not requested_path.is_dir()
        ):
            raise NotADirectoryError(
                f"Pipeline path is not a directory: {pipeline_path}"
            )

        resolved_path = (
            requested_path.resolve()
        )

        if resolved_path.name.endswith(
            PIPELINE_FOLDER_SUFFIX
        ):
            # Canonical format: <output>/<domain_name>_pipelines
            domain_name = resolved_path.name[
                : -len(PIPELINE_FOLDER_SUFFIX)
            ]
            output_base_path = (
                resolved_path.parent
            )
            canonical_path = (
                resolved_path
            )
        elif (
            resolved_path.parent.name
            == LEGACY_ROOT_DIRNAME
        ):
            # Legacy format: <output>/bclearer_pipelines/<domain_name>
            domain_name = resolved_path.name
            output_base_path = (
                resolved_path.parent.parent
            )
            canonical_candidate = (
                _canonical_domain_path(
                    output_base_path,
                    domain_name,
                )
            )
            canonical_path = (
                canonical_candidate
                if canonical_candidate.exists()
                else resolved_path
            )
        else:
            raise ValueError(
                "Path does not follow the expected format "
                f"{LEGACY_ROOT_DIRNAME}/<domain_name> or "
                f"<domain_name>{PIPELINE_FOLDER_SUFFIX}: {pipeline_path}"
            )

        self.requested_path = requested_path
        self.pipeline_path = canonical_path
        self.output_base_path = output_base_path
        self.existing_domain_name = (
            domain_name
        )
        self.legacy_domain_path = (
            _legacy_domain_path(
                self.output_base_path,
                self.existing_domain_name,
            )
        )
        self._b_unit_type = B_UNIT_TYPE_BARE

    def update_pipeline(
        self, config: DomainConfig
    ) -> str:
        """
        Update pipeline structure from configuration.

        Args:
            config: Domain configuration

        Returns:
            Path to the updated pipeline
        """
        self._b_unit_type = config.b_unit_type

        # Verify domain name matches
        if (
            config.domain_name
            != self.existing_domain_name
        ):
            raise ValueError(
                f"Domain name in configuration ('{config.domain_name}') does not match "
                f"existing domain name ('{self.existing_domain_name}')"
            )

        domain_path = self.pipeline_path
        b_source_path = (
            domain_path / "b_source"
        )
        self._create_init_file(
            domain_path
        )

        # Update app_runners/runners directory
        runners_path = (
            b_source_path
            / "app_runners"
            / "runners"
        )

        # Get existing pipelines
        existing_pipelines = self._get_existing_pipelines(
            runners_path
        )

        # Create or update pipelines
        for (
            pipeline_config
        ) in config.pipelines:
            pipeline_name = (
                pipeline_config.name
            )
            pipeline_path = (
                b_source_path
                / pipeline_name
            )

            if (
                pipeline_name
                not in existing_pipelines
            ):
                # Create new pipeline directory if it doesn't exist
                pipeline_path.mkdir(
                    exist_ok=True
                )
                self._create_init_file(
                    pipeline_path
                )

                # Create pipeline subdirectories
                self._create_pipeline_subdirectories(
                    pipeline_path
                )

                # Create pipeline runner
                self._create_pipeline_runner(
                    runners_path,
                    config.domain_name,
                    pipeline_name,
                )

                # Add to existing pipelines list for pipelines_runner update
                existing_pipelines.add(
                    pipeline_name
                )

            # Process thin slices for this pipeline
            self._update_pipeline_thin_slices(
                b_source_path,
                pipeline_config,
                config.domain_name,
            )

        # Update the pipelines_runner to include all pipelines
        self._update_pipelines_runner(
            runners_path,
            config.domain_name,
            list(existing_pipelines),
            config.pipelines,
        )

        _sync_legacy_view(
            self.output_base_path,
            config.domain_name,
            domain_path,
        )

        return str(domain_path)

    def _get_existing_pipelines(
        self, runners_path: Path
    ) -> Set[str]:
        """
        Get names of existing pipelines from runner files.

        Args:
            runners_path: Path to the runners directory

        Returns:
            Set of existing pipeline names
        """
        existing_pipelines = set()

        for (
            file_path
        ) in runners_path.glob(
            "*_runner.py"
        ):
            pipeline_name = (
                file_path.stem.replace(
                    "_runner", ""
                )
            )
            # Exclude pipelines_runner from the list
            if not pipeline_name.endswith(
                "_b_clearer_pipelines"
            ):
                existing_pipelines.add(
                    pipeline_name
                )

        return existing_pipelines

    def _create_pipeline_subdirectories(
        self, pipeline_path: Path
    ) -> None:
        """
        Create subdirectories for a new pipeline.

        Args:
            pipeline_path: Path to the pipeline directory
        """
        # Create objects directory structure
        objects_path = (
            pipeline_path / "objects"
        )
        objects_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            objects_path
        )

        enums_path = (
            objects_path / "enums"
        )
        enums_path.mkdir(exist_ok=True)
        self._create_init_file(
            enums_path
        )

        universes_path = (
            objects_path / "universes"
        )
        universes_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            universes_path
        )

        b_units_path = (
            objects_path / "b_units"
        )
        b_units_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            b_units_path
        )

        # Create operations directory
        operations_path = (
            pipeline_path / "operations"
        )
        operations_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            operations_path
        )

        # Create orchestrators directory structure
        orchestrators_path = (
            pipeline_path
            / "orchestrators"
        )
        orchestrators_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            orchestrators_path
        )

        pipeline_orchestrators_path = (
            orchestrators_path
            / "pipeline"
        )
        pipeline_orchestrators_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            pipeline_orchestrators_path
        )

        thin_slices_path = (
            orchestrators_path
            / "thin_slices"
        )
        thin_slices_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            thin_slices_path
        )

        stages_path = (
            orchestrators_path
            / "stages"
        )
        stages_path.mkdir(exist_ok=True)
        self._create_init_file(
            stages_path
        )

        sub_stages_path = (
            orchestrators_path
            / "sub_stages"
        )
        sub_stages_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            sub_stages_path
        )

    def _update_pipeline_thin_slices(
        self,
        b_source_path: Path,
        pipeline_config: PipelineConfig,
        domain_name: str,
    ) -> None:
        """
        Update or create thin slices for a pipeline.

        Args:
            b_source_path: Path to the b_source directory
            pipeline_config: Pipeline configuration
            domain_name: Name of the domain
        """
        pipeline_name = (
            pipeline_config.name
        )
        pipeline_path = (
            b_source_path
            / pipeline_name
        )

        # Get orchestrators paths
        orchestrators_path = (
            pipeline_path
            / "orchestrators"
        )
        thin_slices_path = (
            orchestrators_path
            / "thin_slices"
        )
        stages_path = (
            orchestrators_path
            / "stages"
        )
        sub_stages_path = (
            orchestrators_path
            / "sub_stages"
        )
        pipeline_orchestrators_path = (
            orchestrators_path
            / "pipeline"
        )

        # Get b_units path
        b_units_path = (
            pipeline_path
            / "objects"
            / "b_units"
        )

        # Create or update pipeline orchestrator
        self._create_pipeline_orchestrator(
            pipeline_orchestrators_path,
            domain_name,
            pipeline_name,
            pipeline_config,
        )

        # Process each thin slice
        for (
            thin_slice
        ) in (
            pipeline_config.thin_slices
        ):
            # Create thin slice orchestrator
            self._create_thin_slice(
                domain_name,
                pipeline_name,
                thin_slice,
                thin_slices_path,
                stages_path,
                sub_stages_path,
                b_units_path,
            )

    def _update_pipelines_runner(
        self,
        runners_path: Path,
        domain_name: str,
        existing_pipelines: List[str],
        new_pipelines: List[
            PipelineConfig
        ],
    ) -> None:
        """
        Update the pipelines runner file to include all pipelines.

        Args:
            runners_path: Path to the runners directory
            domain_name: Name of the domain
            existing_pipelines: List of existing pipeline names
            new_pipelines: List of new pipeline configurations
        """
        # Add any new pipelines that aren't in existing_pipelines yet
        pipeline_names = set(
            existing_pipelines
        )
        for pipeline in new_pipelines:
            pipeline_names.add(
                pipeline.name
            )

        # Sort pipeline names for consistent output
        sorted_pipeline_names = sorted(
            list(pipeline_names)
        )

        runners_path.mkdir(
            parents=True, exist_ok=True
        )
        file_path = (
            runners_path
            / f"{domain_name}_b_clearer_pipelines_runner.py"
        )

        # Create header with imports
        content = PIPELINES_RUNNER_TEMPLATE_HEADER

        # Add imports for each pipeline
        for (
            pipeline_name
        ) in sorted_pipeline_names:
            pipeline_import = PIPELINE_IMPORT_TEMPLATE.format(
                domain_name=domain_name,
                pipeline_name=pipeline_name,
            )
            content += pipeline_import

        # Add function definition
        content += PIPELINES_RUNNER_FUNCTION_START.format(
            domain_name=domain_name
        )

        # Add function calls
        if sorted_pipeline_names:
            for (
                pipeline_name
            ) in sorted_pipeline_names:
                content += PIPELINE_RUNNER_CALL.format(
                    pipeline_name=pipeline_name
                )
        else:
            content += PIPELINES_EMPTY_FUNCTION_BODY

        with open(file_path, "w") as f:
            f.write(content)

    def _create_init_file(
        self, directory_path: Path
    ) -> None:
        """
        Create an empty __init__.py file in a directory.

        Args:
            directory_path: Path to the directory
        """
        init_file = (
            directory_path
            / "__init__.py"
        )
        init_file.touch(exist_ok=True)

    def _create_pipeline_orchestrator(
        self,
        pipeline_orchestrators_path: Path,
        domain_name: str,
        pipeline_name: str,
        pipeline_config: PipelineConfig,
    ) -> None:
        """
        Create pipeline orchestrator file.

        Args:
            pipeline_orchestrators_path: Path to the pipeline orchestrators directory
            domain_name: Name of the domain
            pipeline_name: Name of the pipeline
            pipeline_config: Pipeline configuration
        """
        file_path = (
            pipeline_orchestrators_path
            / f"{pipeline_name}_orchestrator.py"
        )

        content = ""
        # Add imports for each thin slice
        if pipeline_config.thin_slices:
            for (
                thin_slice
            ) in (
                pipeline_config.thin_slices
            ):
                content += PIPELINE_ORCHESTRATOR_IMPORT_TEMPLATE.format(
                    domain_name=domain_name,
                    pipeline_name=pipeline_name,
                    thin_slice_name=thin_slice.name,
                )

        # Add function definition
        content += PIPELINE_ORCHESTRATOR_FUNCTION_START.format(
            pipeline_name=pipeline_name
        )

        # Add calls to thin slice orchestrators
        if pipeline_config.thin_slices:
            for (
                thin_slice
            ) in (
                pipeline_config.thin_slices
            ):
                content += ORCHESTRATE_THIN_SLICE_CALL.format(
                    thin_slice_name=thin_slice.name
                )
        else:
            content += PIPELINE_EMPTY_FUNCTION_BODY

        with open(file_path, "w") as f:
            f.write(content)

    def _create_pipeline_runner(
        self,
        runners_path: Path,
        domain_name: str,
        pipeline_name: str,
    ) -> None:
        """
        Create a pipeline runner file.

        Args:
            runners_path: Path to the runners directory
            domain_name: Name of the domain
            pipeline_name: Name of the pipeline
        """
        file_path = (
            runners_path
            / f"{pipeline_name}_runner.py"
        )

        content = PIPELINE_RUNNER_TEMPLATE.format(
            domain_name=domain_name,
            pipeline_name=pipeline_name,
        )

        with open(file_path, "w") as f:
            f.write(content)

    def _create_thin_slice(
        self,
        domain_name: str,
        pipeline_name: str,
        thin_slice: ThinSliceConfig,
        thin_slices_path: Path,
        stages_path: Path,
        sub_stages_path: Path,
        b_units_path: Path,
    ) -> None:
        """
        Create thin slice orchestrator and related files.

        Args:
            domain_name: Name of the domain
            pipeline_name: Name of the pipeline
            thin_slice: Thin slice configuration
            thin_slices_path: Path to the thin slices directory
            stages_path: Path to the stages directory
            sub_stages_path: Path to the sub-stages directory
            b_units_path: Path to the b_units directory
        """
        slice_name = thin_slice.name
        file_path = (
            thin_slices_path
            / f"{slice_name}_orchestrator.py"
        )

        content = ""
        # Add imports for each stage
        for stage in thin_slice.stages:
            stage_name = stage.name
            content += THIN_SLICE_ORCHESTRATOR_IMPORT_TEMPLATE.format(
                domain_name=domain_name,
                pipeline_name=pipeline_name,
                stage_name=stage_name,
            )

            # Create stage orchestrator
            self._create_stage_orchestrator(
                domain_name,
                pipeline_name,
                stage,
                stages_path,
                sub_stages_path,
                b_units_path,
            )

        # Add function definition
        content += THIN_SLICE_ORCHESTRATOR_FUNCTION_START.format(
            thin_slice_name=slice_name
        )

        # Add calls to stage orchestrators
        if thin_slice.stages:
            for (
                stage
            ) in thin_slice.stages:
                content += ORCHESTRATE_STAGE_CALL.format(
                    pipeline_name=pipeline_name,
                    stage_name=stage.name,
                )
        else:
            content += THIN_SLICE_EMPTY_FUNCTION_BODY

        with open(file_path, "w") as f:
            f.write(content)

    def _create_stage_orchestrator(
        self,
        domain_name: str,
        pipeline_name: str,
        stage: StageConfig,
        stages_path: Path,
        sub_stages_path: Path,
        b_units_path: Path,
    ) -> None:
        """
        Create stage orchestrator file.

        Args:
            domain_name: Name of the domain
            pipeline_name: Name of the pipeline
            stage: Stage configuration
            stages_path: Path to the stages directory
            sub_stages_path: Path to the sub-stages directory
            b_units_path: Path to the b_units directory
        """
        stage_name = stage.name
        file_path = (
            stages_path
            / f"{pipeline_name}_{stage_name}_orchestrator.py"
        )

        # Start with the header imports
        content = (
            STAGE_ORCHESTRATOR_HEADER
        )

        # Add imports for sub-stages if they exist
        if stage.sub_stages:
            for (
                sub_stage
            ) in stage.sub_stages:
                sub_stage_name = (
                    sub_stage.name
                )
                content += SUB_STAGE_IMPORT_TEMPLATE.format(
                    domain_name=domain_name,
                    pipeline_name=pipeline_name,
                    stage_name=stage_name,
                    sub_stage_name=sub_stage_name,
                )

                # Create sub-stage directory and orchestrator
                sub_stage_dir = (
                    sub_stages_path
                    / f"{pipeline_name}_{stage_name}_{sub_stage_name}"
                )
                sub_stage_dir.mkdir(
                    exist_ok=True
                )
                self._create_init_file(
                    sub_stage_dir
                )
                self._create_sub_stage_orchestrator(
                    domain_name,
                    pipeline_name,
                    stage_name,
                    sub_stage,
                    sub_stage_dir,
                    b_units_path,
                )

        # Add b_unit creator import if there are b_units
        if stage.b_units:
            content += STAGE_B_UNIT_CREATOR_IMPORT.format(
                domain_name=domain_name
            )

            # Create stage directory for b_units if doesn't exist
            stage_b_units_dir = (
                b_units_path
                / f"{pipeline_name}_{stage_name}"
            )
            stage_b_units_dir.mkdir(
                exist_ok=True
            )
            self._create_init_file(
                stage_b_units_dir
            )

            # Add imports for each b_unit
            for b_unit in stage.b_units:
                # Format the class name by removing underscores and capitalizing words
                class_name_parts = (
                    b_unit.split("_")
                )
                class_name = "".join(
                    part.capitalize()
                    for part in class_name_parts
                )

                content += B_UNIT_IMPORT_TEMPLATE.format(
                    domain_name=domain_name,
                    pipeline_name=pipeline_name,
                    stage_name=stage_name,
                    b_unit_lower=b_unit.lower(),
                    b_unit_class=class_name,
                )

                # Create b_unit file
                self._create_b_unit(
                    b_unit,
                    stage_b_units_dir,
                )

        # Add function definition
        content += STAGE_ORCHESTRATOR_FUNCTION_START.format(
            pipeline_name=pipeline_name,
            stage_name=stage_name,
        )

        # Add sub-stage calls
        if stage.sub_stages:
            for (
                sub_stage
            ) in stage.sub_stages:
                sub_stage_name = (
                    sub_stage.name
                )
                content += ORCHESTRATE_SUB_STAGE_CALL.format(
                    pipeline_name=pipeline_name,
                    stage_name=stage_name,
                    sub_stage_name=sub_stage_name,
                )

        # Add b_unit calls
        if stage.b_units:
            if stage.sub_stages:
                content += "\n"

            for b_unit in stage.b_units:
                # Format the class name
                class_name_parts = (
                    b_unit.split("_")
                )
                class_name = "".join(
                    part.capitalize()
                    for part in class_name_parts
                )

                content += STAGE_CREATE_AND_RUN_B_UNIT_CALL.format(
                    b_unit_class=class_name,
                )

        # Add empty function body if needed
        if (
            not stage.sub_stages
            and not stage.b_units
        ):
            content += STAGE_EMPTY_FUNCTION_BODY

        with open(file_path, "w") as f:
            f.write(content)

    def _create_sub_stage_orchestrator(
        self,
        domain_name: str,
        pipeline_name: str,
        stage_name: str,
        sub_stage: SubStageConfig,
        sub_stage_dir: Path,
        b_units_path: Path,
    ) -> None:
        """
        Create sub-stage orchestrator file.

        Args:
            domain_name: Name of the domain
            pipeline_name: Name of the pipeline
            stage_name: Name of the stage
            sub_stage: Sub-stage configuration
            sub_stage_dir: Path to the sub-stage directory
            b_units_path: Path to the b_units directory
        """
        sub_stage_name = sub_stage.name
        file_path = (
            sub_stage_dir
            / f"{pipeline_name}_{stage_name}_{sub_stage_name}_orchestrator.py"
        )

        # Start with the header imports
        content = SUB_STAGE_ORCHESTRATOR_HEADER

        # Add b_unit creator import if there are b_units
        if sub_stage.b_units:
            content += SUB_STAGE_B_UNIT_CREATOR_IMPORT.format(
                domain_name=domain_name
            )

            # Create sub-stage directory for b_units if doesn't exist
            sub_stage_b_units_dir = (
                b_units_path
                / f"{pipeline_name}_{stage_name}"
                / f"{pipeline_name}_{stage_name}_{sub_stage_name}"
            )
            sub_stage_b_units_dir.mkdir(
                parents=True,
                exist_ok=True,
            )
            self._create_init_file(
                sub_stage_b_units_dir
            )

            # Add imports for each b_unit
            for (
                b_unit
            ) in sub_stage.b_units:
                # Format the class name
                class_name_parts = (
                    b_unit.split("_")
                )
                class_name = "".join(
                    part.capitalize()
                    for part in class_name_parts
                )

                content += SUB_STAGE_B_UNIT_IMPORT_TEMPLATE.format(
                    domain_name=domain_name,
                    pipeline_name=pipeline_name,
                    stage_name=stage_name,
                    sub_stage_name=sub_stage_name,
                    b_unit_lower=b_unit.lower(),
                    b_unit_class=class_name,
                )

                # Create b_unit file
                self._create_b_unit(
                    b_unit,
                    sub_stage_b_units_dir,
                )

        # Add function definition
        content += SUB_STAGE_ORCHESTRATOR_FUNCTION_START.format(
            pipeline_name=pipeline_name,
            stage_name=stage_name,
            sub_stage_name=sub_stage_name,
        )

        # Add b_unit calls
        if sub_stage.b_units:
            for (
                b_unit
            ) in sub_stage.b_units:
                # Format the class name
                class_name_parts = (
                    b_unit.split("_")
                )
                class_name = "".join(
                    part.capitalize()
                    for part in class_name_parts
                )

                content += SUB_STAGE_CREATE_AND_RUN_B_UNIT_CALL.format(
                    b_unit_class=class_name,
                )
        else:
            content += SUB_STAGE_EMPTY_FUNCTION_BODY

        with open(file_path, "w") as f:
            f.write(content)

    def _get_b_unit_template(self) -> str:
        if self._b_unit_type == B_UNIT_TYPE_WITH_OBJECT_PASSING:
            return B_UNIT_WITH_OBJECT_PASSING_TEMPLATE

        return B_UNIT_TEMPLATE

    def _create_b_unit(
        self,
        b_unit: str,
        directory_path: Path,
    ) -> None:
        """
        Create a b_unit file.

        Args:
            b_unit: Name of the b_unit
            directory_path: Path to the directory for the b_unit
        """
        file_path = (
            directory_path
            / f"{b_unit.lower()}_b_units.py"
        )

        # Skip if the file already exists
        if file_path.exists():
            return

        # Format the class name by removing underscores and capitalizing words
        class_name_parts = b_unit.split(
            "_"
        )
        class_name = "".join(
            part.capitalize()
            for part in class_name_parts
        )

        # Use template to create b_unit file
        content = (
            self._get_b_unit_template().format(
                class_name=class_name
            )
        )

        with open(file_path, "w") as f:
            f.write(content)


class PipelineGenerator:
    """Generator class for creating pipeline structure from configuration."""

    def __init__(
        self,
        output_base_path: str,
    ):
        """
        Initialize the PipelineGenerator.

        Args:
            output_base_path: Base path where the new pipeline will be created
        """
        self.output_base_path = Path(
            output_base_path
        )
        self._b_unit_type = B_UNIT_TYPE_BARE

    def generate_pipeline(
        self, config: DomainConfig
    ) -> str:
        """
        Generate pipeline structure from configuration.

        Args:
            config: Domain configuration

        Returns:
            Path to the generated pipeline
        """
        self._b_unit_type = config.b_unit_type
        domain_name = config.domain_name
        domain_path = _canonical_domain_path(
            self.output_base_path,
            domain_name,
        )
        legacy_path = _legacy_domain_path(
            self.output_base_path,
            domain_name,
        )

        if domain_path.exists():
            raise FileExistsError(
                f"Domain path already exists: {domain_path}"
            )

        if legacy_path.exists():
            raise FileExistsError(
                f"Domain path already exists: {legacy_path}"
            )

        # Create domain directory
        domain_path.mkdir(
            parents=True, exist_ok=True
        )
        self._create_init_file(
            domain_path
        )
        self._create_pyproject(
            domain_path, domain_name
        )

        # Initialize the directory structure
        self._copy_template_structure(
            domain_path
        )

        # Create b_source directory
        b_source_path = (
            domain_path / "b_source"
        )
        b_source_path.mkdir(
            exist_ok=True
        )

        # Create __init__.py in b_source
        self._create_init_file(
            b_source_path
        )

        # Create app_runners directory
        app_runners_path = (
            b_source_path
            / "app_runners"
        )
        app_runners_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            app_runners_path
        )

        # Create aa_b_clearer_pipeline_b_application_runner.py
        self._create_application_runner(
            app_runners_path,
            domain_name,
        )

        # Create app_runners/runners directory
        runners_path = (
            app_runners_path / "runners"
        )
        runners_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            runners_path
        )

        # Create aa_b_clearer_pipelines_runner.py
        self._create_pipelines_runner(
            runners_path,
            domain_name,
            config.pipelines,
        )

        # Create common directory structure
        common_path = (
            b_source_path / "common"
        )
        common_path.mkdir(exist_ok=True)
        self._create_init_file(
            common_path
        )

        common_objects_path = (
            common_path / "objects"
        )
        common_objects_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            common_objects_path
        )

        common_enums_path = (
            common_objects_path
            / "enums"
        )
        common_enums_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            common_enums_path
        )

        common_universes_path = (
            common_objects_path
            / "universes"
        )
        common_universes_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            common_universes_path
        )

        common_operations_path = (
            common_path / "operations"
        )
        common_operations_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            common_operations_path
        )

        b_units_path = (
            common_operations_path
            / "b_units"
        )
        b_units_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            b_units_path
        )

        # Create b_unit_creator_and_runner.py
        self._create_b_unit_creator_and_runner(
            b_units_path
        )

        # Process each pipeline
        for (
            pipeline_config
        ) in config.pipelines:
            self._create_pipeline(
                b_source_path,
                pipeline_config,
                domain_name,
            )

            # Create pipeline runner
            self._create_pipeline_runner(
                runners_path,
                domain_name,
                pipeline_config.name,
            )

        # Create resources directory
        resources_path = (
            domain_path / "resources"
        )
        resources_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            resources_path
        )

        collect_path = (
            resources_path / "collect"
        )
        collect_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            collect_path
        )

        # Create tests directory structure
        tests_path = (
            domain_path / "tests"
        )
        tests_path.mkdir(exist_ok=True)
        self._create_init_file(
            tests_path
        )

        common_tests_path = (
            tests_path / "common"
        )
        common_tests_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            common_tests_path
        )

        fixtures_path = (
            common_tests_path
            / "fixtures"
        )
        fixtures_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            fixtures_path
        )

        inputs_path = (
            common_tests_path / "inputs"
        )
        inputs_path.mkdir(exist_ok=True)
        self._create_init_file(
            inputs_path
        )

        universal_path = (
            tests_path / "universal"
        )
        universal_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            universal_path
        )

        e2e_path = (
            universal_path / "e2e"
        )
        e2e_path.mkdir(exist_ok=True)
        self._create_init_file(e2e_path)

        e2e_fixtures_path = (
            e2e_path / "fixtures"
        )
        e2e_fixtures_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            e2e_fixtures_path
        )

        e2e_outputs_path = (
            e2e_path / "outputs"
        )
        e2e_outputs_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            e2e_outputs_path
        )

        # Create basic test files
        self._create_conftest(e2e_path)
        self._create_e2e_test(
            e2e_path, domain_name
        )
        self._create_startup_script(domain_path)

        _ensure_legacy_view(
            self.output_base_path,
            domain_name,
            domain_path,
        )

        return str(domain_path)

    def _copy_template_structure(
        self, domain_path: Path
    ) -> None:
        """
        Create basic pipeline structure.

        Args:
            domain_path: Path to the domain directory
        """
        # We create all directories and files from scratch for each pipeline
        # This ensures we don't depend on any external template

    def _create_init_file(
        self, directory_path: Path
    ) -> None:
        """
        Create an empty __init__.py file in a directory.

        Args:
            directory_path: Path to the directory
        """
        init_file = (
            directory_path
            / "__init__.py"
        )
        init_file.touch()

    def _create_application_runner(
        self,
        app_runners_path: Path,
        domain_name: str,
    ) -> None:
        """
        Create the main application runner file.

        Args:
            app_runners_path: Path to the app_runners directory
            domain_name: Name of the domain
        """
        file_path = (
            app_runners_path
            / f"{domain_name}_b_clearer_pipeline_b_application_runner.py"
        )

        content = APPLICATION_RUNNER_TEMPLATE.format(
            domain_name=domain_name
        )

        with open(file_path, "w") as f:
            f.write(content)

    def _create_pipelines_runner(
        self,
        runners_path: Path,
        domain_name: str,
        pipelines: List[PipelineConfig],
    ) -> None:
        """
        Create the pipelines runner file.

        Args:
            runners_path: Path to the runners directory
            domain_name: Name of the domain
            pipelines: List of pipeline configurations
        """
        runners_path.mkdir(
            parents=True, exist_ok=True
        )
        file_path = (
            runners_path
            / f"{domain_name}_b_clearer_pipelines_runner.py"
        )

        # Start with the header imports
        content = PIPELINES_RUNNER_TEMPLATE_HEADER

        # Add imports for each pipeline
        for pipeline in pipelines:
            content += PIPELINE_IMPORT_TEMPLATE.format(
                domain_name=domain_name,
                pipeline_name=pipeline.name,
            )

        # Add function definition
        content += PIPELINES_RUNNER_FUNCTION_START.format(
            domain_name=domain_name
        )

        # Add function calls
        if pipelines:
            for pipeline in pipelines:
                content += PIPELINE_RUNNER_CALL.format(
                    pipeline_name=pipeline.name
                )
        else:
            content += PIPELINES_EMPTY_FUNCTION_BODY

        with open(file_path, "w") as f:
            f.write(content)

    def _create_pipeline_runner(
        self,
        runners_path: Path,
        domain_name: str,
        pipeline_name: str,
    ) -> None:
        """
        Create a pipeline runner file.

        Args:
            runners_path: Path to the runners directory
            domain_name: Name of the domain
            pipeline_name: Name of the pipeline
        """
        runners_path.mkdir(
            parents=True, exist_ok=True
        )
        file_path = (
            runners_path
            / f"{pipeline_name}_runner.py"
        )

        content = PIPELINE_RUNNER_TEMPLATE.format(
            domain_name=domain_name,
            pipeline_name=pipeline_name,
        )

        with open(file_path, "w") as f:
            f.write(content)

    def _create_pipeline(
        self,
        b_source_path: Path,
        pipeline_config: PipelineConfig,
        domain_name: str,
    ) -> None:
        """
        Create a pipeline directory structure.

        Args:
            b_source_path: Path to the b_source directory
            pipeline_config: Pipeline configuration
            domain_name: Name of the domain
        """
        pipeline_name = (
            pipeline_config.name
        )
        pipeline_path = (
            b_source_path
            / pipeline_name
        )
        pipeline_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            pipeline_path
        )

        # Create pipeline objects directory
        objects_path = (
            pipeline_path / "objects"
        )
        objects_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            objects_path
        )

        enums_path = (
            objects_path / "enums"
        )
        enums_path.mkdir(exist_ok=True)
        self._create_init_file(
            enums_path
        )

        universes_path = (
            objects_path / "universes"
        )
        universes_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            universes_path
        )

        b_units_path = (
            objects_path / "b_units"
        )
        b_units_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            b_units_path
        )

        # Create operations directory
        operations_path = (
            pipeline_path / "operations"
        )
        operations_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            operations_path
        )

        # Create orchestrators directory
        orchestrators_path = (
            pipeline_path
            / "orchestrators"
        )
        orchestrators_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            orchestrators_path
        )

        pipeline_orchestrators_path = (
            orchestrators_path
            / "pipeline"
        )
        pipeline_orchestrators_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            pipeline_orchestrators_path
        )

        # Create thin_slices directory
        thin_slices_path = (
            orchestrators_path
            / "thin_slices"
        )
        thin_slices_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            thin_slices_path
        )

        # Create stages directory
        stages_path = (
            orchestrators_path
            / "stages"
        )
        stages_path.mkdir(exist_ok=True)
        self._create_init_file(
            stages_path
        )

        # Create sub_stages directory
        sub_stages_path = (
            orchestrators_path
            / "sub_stages"
        )
        sub_stages_path.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            sub_stages_path
        )

        # Create pipeline orchestrator
        self._create_pipeline_orchestrator(
            pipeline_orchestrators_path,
            domain_name,
            pipeline_name,
            pipeline_config,
        )

        # Create thin slices
        for (
            thin_slice
        ) in (
            pipeline_config.thin_slices
        ):
            self._create_thin_slice(
                domain_name,
                pipeline_name,
                thin_slice,
                thin_slices_path,
                stages_path,
                sub_stages_path,
                b_units_path,
            )

    def _create_pipeline_orchestrator(
        self,
        pipeline_orchestrators_path: Path,
        domain_name: str,
        pipeline_name: str,
        pipeline_config: PipelineConfig,
    ) -> None:
        """
        Create pipeline orchestrator file.

        Args:
            pipeline_orchestrators_path: Path to the pipeline orchestrators directory
            domain_name: Name of the domain
            pipeline_name: Name of the pipeline
            pipeline_config: Pipeline configuration
        """
        file_path = (
            pipeline_orchestrators_path
            / f"{pipeline_name}_orchestrator.py"
        )

        content = ""
        # Add imports for each thin slice
        if pipeline_config.thin_slices:
            for (
                thin_slice
            ) in (
                pipeline_config.thin_slices
            ):
                content += PIPELINE_ORCHESTRATOR_IMPORT_TEMPLATE.format(
                    domain_name=domain_name,
                    pipeline_name=pipeline_name,
                    thin_slice_name=thin_slice.name,
                )

        # Add function definition
        content += PIPELINE_ORCHESTRATOR_FUNCTION_START.format(
            pipeline_name=pipeline_name
        )

        # Add calls to thin slice orchestrators
        if pipeline_config.thin_slices:
            for (
                thin_slice
            ) in (
                pipeline_config.thin_slices
            ):
                content += ORCHESTRATE_THIN_SLICE_CALL.format(
                    thin_slice_name=thin_slice.name
                )
        else:
            content += PIPELINE_EMPTY_FUNCTION_BODY

        with open(file_path, "w") as f:
            f.write(content)

    def _create_thin_slice(
        self,
        domain_name: str,
        pipeline_name: str,
        thin_slice: ThinSliceConfig,
        thin_slices_path: Path,
        stages_path: Path,
        sub_stages_path: Path,
        b_units_path: Path,
    ) -> None:
        """
        Create thin slice orchestrator and related files.

        Args:
            domain_name: Name of the domain
            pipeline_name: Name of the pipeline
            thin_slice: Thin slice configuration
            thin_slices_path: Path to the thin slices directory
            stages_path: Path to the stages directory
            sub_stages_path: Path to the sub-stages directory
            b_units_path: Path to the b_units directory
        """
        slice_name = thin_slice.name
        file_path = (
            thin_slices_path
            / f"{slice_name}_orchestrator.py"
        )

        content = ""
        # Add imports for each stage
        for stage in thin_slice.stages:
            stage_name = stage.name
            content += THIN_SLICE_ORCHESTRATOR_IMPORT_TEMPLATE.format(
                domain_name=domain_name,
                pipeline_name=pipeline_name,
                stage_name=stage_name,
            )

            # Create stage orchestrator
            self._create_stage_orchestrator(
                domain_name,
                pipeline_name,
                stage,
                stages_path,
                sub_stages_path,
                b_units_path,
            )

        # Add function definition
        content += THIN_SLICE_ORCHESTRATOR_FUNCTION_START.format(
            thin_slice_name=slice_name
        )

        # Add calls to stage orchestrators
        if thin_slice.stages:
            for (
                stage
            ) in thin_slice.stages:
                content += ORCHESTRATE_STAGE_CALL.format(
                    pipeline_name=pipeline_name,
                    stage_name=stage.name,
                )
        else:
            content += THIN_SLICE_EMPTY_FUNCTION_BODY

        with open(file_path, "w") as f:
            f.write(content)

    def _create_stage_orchestrator(
        self,
        domain_name: str,
        pipeline_name: str,
        stage: StageConfig,
        stages_path: Path,
        sub_stages_path: Path,
        b_units_path: Path,
    ) -> None:
        """
        Create stage orchestrator file.

        Args:
            domain_name: Name of the domain
            pipeline_name: Name of the pipeline
            stage: Stage configuration
            stages_path: Path to the stages directory
            sub_stages_path: Path to the sub-stages directory
            b_units_path: Path to the b_units directory
        """
        stage_name = stage.name
        file_path = (
            stages_path
            / f"{pipeline_name}_{stage_name}_orchestrator.py"
        )

        # Always create stage directory for b_units and __init__.py file
        stage_b_units_dir = (
            b_units_path
            / f"{pipeline_name}_{stage_name}"
        )
        stage_b_units_dir.mkdir(
            exist_ok=True
        )
        self._create_init_file(
            stage_b_units_dir
        )

        # Start with the header imports
        content = (
            STAGE_ORCHESTRATOR_HEADER
        )

        # Add imports for sub-stages if they exist
        if stage.sub_stages:
            for (
                sub_stage
            ) in stage.sub_stages:
                sub_stage_name = (
                    sub_stage.name
                )
                content += SUB_STAGE_IMPORT_TEMPLATE.format(
                    domain_name=domain_name,
                    pipeline_name=pipeline_name,
                    stage_name=stage_name,
                    sub_stage_name=sub_stage_name,
                )

                # Create sub-stage directory and orchestrator
                sub_stage_dir = (
                    sub_stages_path
                    / f"{pipeline_name}_{stage_name}_{sub_stage_name}"
                )
                sub_stage_dir.mkdir(
                    exist_ok=True
                )
                self._create_init_file(
                    sub_stage_dir
                )
                self._create_sub_stage_orchestrator(
                    domain_name,
                    pipeline_name,
                    stage_name,
                    sub_stage,
                    sub_stage_dir,
                    b_units_path,
                )

        # Add b_unit creator import if there are b_units
        if stage.b_units:
            content += STAGE_B_UNIT_CREATOR_IMPORT.format(
                domain_name=domain_name
            )

            for b_unit in stage.b_units:
                # Format the class name by removing underscores and capitalizing words
                class_name_parts = (
                    b_unit.split("_")
                )
                class_name = "".join(
                    part.capitalize()
                    for part in class_name_parts
                )

                content += B_UNIT_IMPORT_TEMPLATE.format(
                    domain_name=domain_name,
                    pipeline_name=pipeline_name,
                    stage_name=stage_name,
                    b_unit_lower=b_unit.lower(),
                    b_unit_class=class_name,
                )

                # Create b_unit file
                self._create_b_unit(
                    b_unit,
                    stage_b_units_dir,
                )

        # Add function definition
        content += STAGE_ORCHESTRATOR_FUNCTION_START.format(
            pipeline_name=pipeline_name,
            stage_name=stage_name,
        )

        # Add sub-stage calls
        if stage.sub_stages:
            for (
                sub_stage
            ) in stage.sub_stages:
                sub_stage_name = (
                    sub_stage.name
                )
                content += ORCHESTRATE_SUB_STAGE_CALL.format(
                    pipeline_name=pipeline_name,
                    stage_name=stage_name,
                    sub_stage_name=sub_stage_name,
                )

        # Add b_unit calls
        if stage.b_units:
            if stage.sub_stages:
                content += "\n"

            for b_unit in stage.b_units:
                # Format the class name
                class_name_parts = (
                    b_unit.split("_")
                )
                class_name = "".join(
                    part.capitalize()
                    for part in class_name_parts
                )

                content += STAGE_CREATE_AND_RUN_B_UNIT_CALL.format(
                    b_unit_class=class_name,
                )

        # Add empty function body if needed
        if (
            not stage.sub_stages
            and not stage.b_units
        ):
            content += STAGE_EMPTY_FUNCTION_BODY

        with open(file_path, "w") as f:
            f.write(content)

    def _create_sub_stage_orchestrator(
        self,
        domain_name: str,
        pipeline_name: str,
        stage_name: str,
        sub_stage: SubStageConfig,
        sub_stage_dir: Path,
        b_units_path: Path,
    ) -> None:
        """
        Create sub-stage orchestrator file.

        Args:
            domain_name: Name of the domain
            pipeline_name: Name of the pipeline
            stage_name: Name of the stage
            sub_stage: Sub-stage configuration
            sub_stage_dir: Path to the sub-stage directory
            b_units_path: Path to the b_units directory
        """
        sub_stage_name = sub_stage.name
        file_path = (
            sub_stage_dir
            / f"{pipeline_name}_{stage_name}_{sub_stage_name}_orchestrator.py"
        )

        # Start with the header imports
        content = SUB_STAGE_ORCHESTRATOR_HEADER

        # Add b_unit creator import if there are b_units
        if sub_stage.b_units:
            content += SUB_STAGE_B_UNIT_CREATOR_IMPORT.format(
                domain_name=domain_name
            )

            # Create sub-stage directory for b_units if doesn't exist
            sub_stage_b_units_dir = (
                b_units_path
                / f"{pipeline_name}_{stage_name}"
                / f"{pipeline_name}_{stage_name}_{sub_stage_name}"
            )
            sub_stage_b_units_dir.mkdir(
                parents=True,
                exist_ok=True,
            )
            self._create_init_file(
                sub_stage_b_units_dir
            )

            # Add imports for each b_unit
            for (
                b_unit
            ) in sub_stage.b_units:
                # Format the class name
                class_name_parts = (
                    b_unit.split("_")
                )
                class_name = "".join(
                    part.capitalize()
                    for part in class_name_parts
                )

                content += SUB_STAGE_B_UNIT_IMPORT_TEMPLATE.format(
                    domain_name=domain_name,
                    pipeline_name=pipeline_name,
                    stage_name=stage_name,
                    sub_stage_name=sub_stage_name,
                    b_unit_lower=b_unit.lower(),
                    b_unit_class=class_name,
                )

                # Create b_unit file
                self._create_b_unit(
                    b_unit,
                    sub_stage_b_units_dir,
                )

        # Add function definition
        content += SUB_STAGE_ORCHESTRATOR_FUNCTION_START.format(
            pipeline_name=pipeline_name,
            stage_name=stage_name,
            sub_stage_name=sub_stage_name,
        )

        # Add b_unit calls
        if sub_stage.b_units:
            for (
                b_unit
            ) in sub_stage.b_units:
                # Format the class name
                class_name_parts = (
                    b_unit.split("_")
                )
                class_name = "".join(
                    part.capitalize()
                    for part in class_name_parts
                )

                content += SUB_STAGE_CREATE_AND_RUN_B_UNIT_CALL.format(
                    b_unit_class=class_name,
                )
        else:
            content += SUB_STAGE_EMPTY_FUNCTION_BODY

        with open(file_path, "w") as f:
            f.write(content)

    def _get_b_unit_template(self) -> str:
        if self._b_unit_type == B_UNIT_TYPE_WITH_OBJECT_PASSING:
            return B_UNIT_WITH_OBJECT_PASSING_TEMPLATE

        return B_UNIT_TEMPLATE

    def _create_b_unit(
        self,
        b_unit: str,
        directory_path: Path,
    ) -> None:
        """
        Create a b_unit file.

        Args:
            b_unit: Name of the b_unit
            directory_path: Path to the directory for the b_unit
        """
        file_path = (
            directory_path
            / f"{b_unit.lower()}_b_units.py"
        )

        # Skip if the file already exists
        if file_path.exists():
            return

        # Format the class name by removing underscores and capitalizing words
        class_name_parts = b_unit.split(
            "_"
        )
        class_name = "".join(
            part.capitalize()
            for part in class_name_parts
        )

        # Use template to create b_unit file
        content = (
            self._get_b_unit_template().format(
                class_name=class_name
            )
        )

        with open(file_path, "w") as f:
            f.write(content)

    def _create_b_unit_creator_and_runner(
        self, b_units_path: Path
    ) -> None:
        """
        Create b_unit_creator_and_runner.py file.

        Args:
            b_units_path: Path to the b_units directory
        """
        file_path = (
            b_units_path
            / "b_unit_creator_and_runner.py"
        )

        with open(file_path, "w") as f:
            f.write(
                B_UNIT_CREATOR_AND_RUNNER_TEMPLATE
            )

    def _create_conftest(
        self, e2e_path: Path
    ) -> None:
        """
        Create a basic conftest.py file for tests.

        Args:
            e2e_path: Path to the e2e directory
        """
        file_path = (
            e2e_path / "conftest.py"
        )

        with open(file_path, "w") as f:
            f.write(CONFTEST_TEMPLATE)

    def _create_e2e_test(
        self,
        e2e_path: Path,
        domain_name: str,
    ) -> None:
        """
        Create a basic e2e test file.

        Args:
            e2e_path: Path to the e2e directory
            domain_name: Name of the domain
        """
        file_path = (
            e2e_path
            / f"test_{domain_name}_b_clearer_pipeline_b_application_runner.py"
        )

        content = (
            E2E_TEST_TEMPLATE.format(
                domain_name=domain_name
            )
        )

        with open(file_path, "w") as f:
            f.write(content)

    def _create_pyproject(
        self,
        domain_path: Path,
        domain_name: str,
    ) -> None:
        file_path = domain_path / "pyproject.toml"

        with open(file_path, "w") as f:
            f.write(
                PYPROJECT_TEMPLATE.format(
                    domain_name=domain_name
                )
            )

    def _create_startup_script(
        self, domain_path: Path
    ) -> None:
        file_path = domain_path / "run_e2e.sh"

        with open(file_path, "w") as f:
            f.write(STARTUP_SCRIPT_TEMPLATE)

        os.chmod(file_path, 0o755)


def generate_pipeline(
    config: Dict,
    output_base_path: str,
) -> str:
    """
    Generate a pipeline from configuration.

    Args:
        config: Pipeline configuration dictionary
        output_base_path: Base path where the new pipeline will be created

    Returns:
        Path to the generated pipeline
    """
    from bclearer_core.pipeline_builder.schema import (
        validate_pipeline_config,
    )

    config_obj = (
        validate_pipeline_config(config)
    )
    generator = PipelineGenerator(
        output_base_path
    )

    return generator.generate_pipeline(
        config_obj
    )


def update_pipeline(
    config: Dict, pipeline_path: str
) -> str:
    """
    Update an existing pipeline with new configuration.

    Args:
        config: Pipeline configuration dictionary
        pipeline_path: Path to the existing pipeline

    Returns:
        Path to the updated pipeline
    """
    from bclearer_core.pipeline_builder.schema import (
        validate_pipeline_config,
    )

    config_obj = (
        validate_pipeline_config(config)
    )
    updater = PipelineUpdater(
        pipeline_path
    )
    return updater.update_pipeline(
        config_obj
    )



"""Template for pyproject.toml."""

PYPROJECT_TEMPLATE = """[project]
name = "{domain_name}_pipelines"
version = "0.1.0"
description = "Generated bclearer pipeline package"
requires-python = ">=3.10"
dependencies = [
    "pytest",
    "psutil",
]

[build-system]
requires = ["setuptools>=61.0"]
build-backend = "setuptools.build_meta"
"""



import datetime

import psutil


def get_user_session_start_date_time(
    user_name: str,
) -> datetime.datetime:
    user = __get_user_by_name(
        user_name=user_name
    )

    if not user:
        return datetime.datetime.fromtimestamp(
            psutil.boot_time()
        )

    user_session_start_date_time = (
        user.started
    )

    user_session_start_date_time_formatted = datetime.datetime.fromtimestamp(
        user_session_start_date_time
    )

    return user_session_start_date_time_formatted


def __get_user_by_name(user_name: str):
    for user in psutil.users():
        if user.name == user_name:
            return user

    return None



import getpass
import os
from typing import Tuple

from bclearer_core.infrastructure.session.common.enums.operating_system_kind import (
    OperatingSystemKind,
)
from bclearer_core.infrastructure.session.common.platform_type_getter import (
    get_operating_system_kind,
)
from bclearer_core.infrastructure.session.common.session_identity_vectors import (
    OsUserProfileObjectIdentityVector,
    OsUserProfileObjectIdentityVectorPlaces,
)
from bclearer_core.infrastructure.session.common.session_objects import (
    SessionObjects,
)
from bclearer_core.infrastructure.session.operating_system.operating_system_objects import (
    OperatingSystemObjects,
)
from bclearer_core.infrastructure.session.user.bie_os_user_profile_types import (
    BieOsUserProfileTypes,
)
from bclearer_core.infrastructure.session.user.helpers.bie_os_user_profile_type_getter import (
    get_bie_os_user_profile_type,
)
from bclearer_core.infrastructure.session.user.helpers.unix.unix_account_context import (
    get_unix_account_context,
)
from bclearer_core.infrastructure.session.user.helpers.user_sid_getter import (
    get_user_sid,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.bie_id_creation_module.bie_id_creation_facade import (
    BieIdCreationFacade,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.objects.bie_ids import (
    BieIds,
)


class OsUserProfileObjects(
    SessionObjects
):
    _instance = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(
                OsUserProfileObjects,
                cls,
            ).__new__(cls)

        return cls._instance

    def __init__(self):
        self.user_name = (
            self.__get_user_name()
        )

        self.user_sid = get_user_sid()

        (
            self.domain_name,
            self.account_type,
        ) = (
            self.__resolve_account_context()
        )

        os_user_profile_type = (
            get_bie_os_user_profile_type()
        )

        bie_id = self.__create_bie_id(
            os_user_profile_type=os_user_profile_type
        )

        super().__init__(
            bie_id=bie_id,
            base_hr_name=self.user_name,
            bie_infrastructure_type=os_user_profile_type,
        )

    @staticmethod
    def __get_user_name() -> str:
        try:
            return getpass.getuser()
        except Exception:
            return (
                os.environ.get(
                    "USERNAME"
                )
                or os.environ.get(
                    "USER"
                )
                or "unknown-user"
            )

    def __resolve_account_context(
        self,
    ) -> Tuple[str, str]:
        operating_system_kind = (
            get_operating_system_kind()
        )

        if (
            operating_system_kind
            is OperatingSystemKind.WINDOWS
        ):
            from bclearer_core.infrastructure.session.user.helpers.windows.windows_account_context import (
                get_windows_account_context,
            )

            return get_windows_account_context(
                self.user_sid
            )

        return (
            get_unix_account_context()
        )

    def __create_bie_id(
        self,
        os_user_profile_type: BieOsUserProfileTypes,
    ) -> BieIds:
        operating_system_object = (
            OperatingSystemObjects()
        )

        identity_vector = OsUserProfileObjectIdentityVector(
            places=OsUserProfileObjectIdentityVectorPlaces(
                os_user_profile_type_bie_id=os_user_profile_type.item_bie_identity,
                user_sid=self.user_sid,
                operating_system_type_bie_id=operating_system_object.bie_operating_system_type.item_bie_identity,
            )
        )

        user_base_bie_id = BieIdCreationFacade.create_bie_id_from_identity_vector(
            identity_vector=identity_vector,
        )

        return user_base_bie_id



"""Template for b_unit creator and runner."""

B_UNIT_CREATOR_AND_RUNNER_TEMPLATE = """def create_and_run_b_unit(
    b_unit_type,
    input_object=None,
) -> None:
    if input_object is None:
        b_unit = b_unit_type()
    else:
        b_unit = b_unit_type(
            input_object=input_object
        )

    b_unit.run()
"""



from enum import auto

from bclearer_orchestration_services.identification_services.b_identity_ecosystem.common_knowledge.bie_enums import (
    BieEnums,
)


class OsUserProfileObjectColumnNames(
    BieEnums
):
    NOT_SET = auto()

    USER_NAMES = auto()

    USER_SIDS = auto()

    DOMAIN_NAMES = auto()

    ACCOUNT_TYPES = auto()



import os

WRITE_ACRONYM = "w"

UTF_8_ENCODING_NAME = "utf-8"

DEFAULT_NULL_VALUE = "DefaultNullValue"

FOLDER_DELIMITER = os.sep

CLONED_DATA_FOLDER_NAME = "cloned_data"

LOCAL_GIT_REPOSITORY_FOLDER_NAME = (
    "local_git_repositories"
)

LOCAL_CONFIGURATION_FOLDER = (
    "local_configuration"
)

LOCAL_CONFIGURATION_FILE = (
    "local_configuration.json_service"
)

ORIGINAL_DATA_CONFIGURATION_FILE = "original_data_configuration.json_service"

CODE_FOLDER_NAME = "code"

SANDPIT_FOLDER_NAME = "sandpit"

EXECUTABLES_FOLDER_NAME = "executables"

PYTHON_COMMIT_MESSAGE = (
    "Python Commit - Code and CD"
)

B_UNIT_CONFIGURATION_FILE = (
    "b_unit_configuration.json_service"
)

PROJECT_SECTION_B_UNITS = "b_units"

B_UNIT_ATTRIBUTE_B_UNIT_NAME = "name"

B_UNIT_SECTION_INPUTS = "inputs"

INPUT_ATTRIBUTE_GIT_REPOSITORY = (
    "git_repository"
)

INPUT_ATTRIBUTE_GIT_RELEASE_GUID = (
    "release_guid"
)

INPUT_ATTRIBUTE_FILE_NAME = "file_name"

CHECK_RESULTS_DATAFRAME_NAME_SUCCESS_SUFFIX = (
    "success"
)

CHECK_RESULTS_DATAFRAME_NAME_FAILURE_SUFFIX = (
    "failure"
)

NAME_COMPONENT_DIVIDER = "_"

HR_NAME_COMPONENT_DIVIDER = "_"

HR_DATE_TIME_FORMAT = "%Y%m%d_%H%M"

DEFAULT_MASTER_TABLE_SUFFIX = "_1"

DEFAULT_FOREIGN_TABLE_SUFFIX = "_2"
